"""Utility functions for downloading agent files."""
import os
import os.path as osp
import pathlib

import requests
from clint.textui import progress  # type: ignore

from posggym import error, logger
from posggym.config import AGENT_MODEL_REPO_URL

# largest policy file is ~ 1.3-4 MB
LARGEST_FILE_SIZE = int(1.5 * 1024 * 1024)


def download_to_file(url: str, dest_file_path: str):
    """Download file from URL and store at specified destination.

    Arguments
    ---------
    url: full url to download file from
    dest_file_path: file path to write downloaded file to.

    Raises
    ------
    posggym.error.DownloadError: if error occurred while trying to download file.

    """
    dest_dir = osp.dirname(dest_file_path)
    if not osp.exists(dest_dir):
        # create dir if it does not exist
        os.makedirs(dest_dir)

    r = requests.get(url, stream=True)
    if r.ok:
        with open(dest_file_path, "wb") as f:
            content_len = r.headers.get("content-length")
            if isinstance(content_len, str):
                try:
                    total_length = int(content_len)
                except (TypeError,):
                    total_length = LARGEST_FILE_SIZE
            else:
                total_length = LARGEST_FILE_SIZE

            for chunk in progress.bar(
                r.iter_content(chunk_size=1024), expected_size=(total_length / 1024) + 1
            ):
                if chunk:
                    f.write(chunk)
                    f.flush()
                    os.fsync(f.fileno())

    else:
        # HTTP status code 4XX/5XX
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            # wrap exception in posggym-agents error
            raise error.DownloadError(
                f"Error while downloading file, caused by: {type(e).__name__}: {str(e)}"
            ) from e


def download_from_repo(file_path: str, rewrite_existing: bool = False):
    """Download file from the posggym-agent-models github repo.

    Arguments
    ---------
    file_path: local path to posgym package file.
    rewrite_existing: whether to re-download and rewrite an existing copy of the file.

    Raises
    ------
    posggym.error.InvalidFile: if file_path is not a valid posggym-agents package file.
    posggym.error.DownloadError: if error occurred while trying to download file.

    """
    if osp.exists(file_path) and not rewrite_existing:
        return

    path = pathlib.Path(file_path)
    if "agents" not in path.parts:
        raise error.InvalidFile(
            f"Invalid posggym.agents file path '{file_path}'. Path must contain the "
            "`agents` directory."
        )

    base_repo_dir_index = path.parts.index("agents")
    file_repo_url = (
        AGENT_MODEL_REPO_URL + "posggym/" + "/".join(path.parts[base_repo_dir_index:])
    )

    logger.info(
        f"Downloading file from posggym-agent-models repository: {file_repo_url}."
    )

    return download_to_file(file_repo_url, file_path)



"""Contains the main POSG environment class and functions.

The implementation is heavily inspired by Open AI Gym
https://github.com/openai/gym

And, the more recent, Farama Foundation Gymnasium
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/core.py

"""
from __future__ import annotations

import abc
import copy
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Optional, Tuple, TypeVar

from posggym.model import ActType, ObsType, POSGModel, StateType

if TYPE_CHECKING:
    import numpy as np
    from gymnasium import spaces

    from posggym.envs.registration import EnvSpec


class Env(abc.ABC, Generic[StateType, ObsType, ActType]):
    r"""The main POSGGym class for implementing POSG environments.

    The class encapsulates an environment and a POSG model. The environment maintains an
    internal state and can be interacted with by multiple agents in parallel through
    the :meth:`step` and :meth:`reset` functions. The POSG model can be accessed via the
    :attr:`model` attribute and exposes the model of the environment which can be used
    for planning or for anything else (see :py:class:`posggym.POSGModel` class for
    details).

    The implementation is heavily inspired by the Farama Foundation Gymnasium
    (https://github.com/Farama-Foundation/Gymnasium) and PettingZoo
    (https://github.com/Farama-Foundation/PettingZoo) APIs. It aims to be consistent
    with these APIs and easily compatible with the PettingZoo API.

    The main API methods that users of this class need to know are:

    - :meth:`step`
    - :meth:`reset`
    - :meth:`render`
    - :meth:`close`

    And the main attributes:

    - :attr:`model` - The POSG model of the environment (:py:class:`posggym.POSGModel`)
    - :attr:`state` - The current state of the environment
    - :attr:`possible_agents` - All agents that may appear in the environment
    - :attr:`agents` - The agents currently active in the environment
    - :attr:`action_spaces` - The action space for each agent
    - :attr:`observation_spaces` - The observation space for each agent
    - :attr:`reward_ranges` - The minimum and maximum possible rewards each agent may
      receive for single step in the environment. The default reward range is set to
      :math:`(-\infty,+\infty)`.
    - :attr:`is_symmetric` - Whether the environment is symmetric or asymmetric
    - :attr:`spec` - An environment spec that contains the information used to
      initialize the environment from :meth:`posggym.make`
    - :attr:`metadata` - The metadata of the environment, i.e. render modes, render fps
    - :attr:`render_mode` - The current render mode of the environment

    Environments have additional methods and attributes that provide more environment
    information and access:

    - :attr:`unwrapped`

    """

    # Set this in SOME subclasses
    metadata: Dict[str, Any] = {"render_modes": []}

    # Define render_mode if your environment supports rendering
    render_mode: str | None = None

    # EnvSpec used to instantiate env instance
    # This is set when env is made using posggym.make function
    spec: EnvSpec | None = None

    # The model used by the environment
    model: POSGModel[StateType, ObsType, ActType]

    @abc.abstractmethod
    def step(
        self, actions: Dict[str, ActType]
    ) -> Tuple[
        Dict[str, ObsType],
        Dict[str, float],
        Dict[str, bool],
        Dict[str, bool],
        bool,
        Dict[str, Dict[str, Any]],
    ]:
        """Run one timestep in the environment using the agents' actions.

        When the end of an episode is reached, the user is responsible for
        calling :meth:`reset()` to reset this environments state.

        Arguments
        ---------
        actions : Dict[str, ActType]
          a joint action containing one action per active agent in the environment.

        Returns
        -------
        observations : Dict[str, ObsType]
          the joint observation containing one observation per agent.
        rewards : Dict[str, float]
          the joint rewards containing one reward per agent.
        terminations : Dict[str, bool]
          whether each agent has reached a terminal state in the environment.
          Contains one value for each agent in the environment. It's possible,
          depending on the environment, for only some of the agents to be in a
          terminal during a given step.
        truncations : Dict[str, bool]
          whether the episode has been truncated for each agent in the
          environment. Contains one value for each agent in the environment. Truncation
          for an agent signifies that the episode was ended for that agent (e.g. due to
          reaching the time limit) before the agent reached a terminal state.
        all_done : bool
          whether the episode is finished. Provided for convenience and to handle
          the case where agents may be added and removed during an episode. For
          environments where the active agents remains constant during each episode,
          this is equivalent to checking if all agents are either in a terminated or
          truncated state. If true, the user needs to call :py:func:`reset()`.
        infos : Dict[str, Dict[str, Any]]
          contains auxiliary diagnostic information (helpful for debugging, learning,
          and logging) for each agent.

        """

    def reset(
        self, *, seed: int | None = None, options: Dict[str, Any] | None = None
    ) -> Tuple[Dict[str, ObsType], Dict[str, Dict]]:
        """Resets the environment and returns an initial observations and info.

        This method generates a new starting state often with some randomness. This
        randomness can be controlled with the ``seed`` parameter. If the
        environment already has a random number generator (RNG) and :meth:`reset` is
        called with ``seed=None``, the RNG is not reset. Note, that the RNG is handled
        by the environment :attr:`model`, rather than the environment class itself.

        Therefore, :meth:`reset` should (in the typical use case) be called with a seed
        right after initialization and then never again.

        For Custom environments, the first line of :meth:`reset` should be
        ``super().reset(seed=seed)`` which implements the seeding correctly.

        Arguments
        ---------
        seed : int,  optional
          The seed that is used to initialize the environment's RNG. If the
          ``seed=None`` is passed, the RNG will *not* be reset. If you pass an
          integer, the RNG will be reset even if it already exists. Usually, you want
          to pass an integer *right after the environment has been initialized and
          then never again*.
        options: dict, optional
          Additional information to specify how the environment is reset (optional,
          depending on the specific environment)

        Returns
        -------
        observations : Dict[str, ObsType], optional
          The joint observation containing one observation per agent in the environment.
        infos : Dict[str, Dict]
          Auxiliary information for each agent. It should be analogous to the ``info``
          returned by :meth:`step()` and can be empty.

        """
        # initialize the RNG if the seed is manually passed
        if seed is not None:
            self.model.seed(seed)
        return {}, {}

    def render(
        self,
    ) -> None | np.ndarray | str | Dict[str, np.ndarray] | Dict[str, str]:
        """Render the environment as specified by environment :attr:`render_mode`.

        The render mode attribute :attr:`render_mode` is set during the initialization
        of the environment. While the environment's :attr:`metadata` render modes
        (`env.metadata["render_modes"]`) should contain the supported render modes.

        The set of supported modes varies per environment (some environments do not
        support rendering at all). By convention, if :attr:`render_mode` is:

        - ``None`` (default): no render is computed.
        - ``"human"``: Environment is rendered to the current display or terminal
          usually for human consumption. Returns ``None``.
        - ``"rgb_array"``: Return an ``np.ndarray`` with shape ``(x, y, 3)``
          representing RGB values for an x-by-y pixel image of the entire environment,
          suitable for turning into a video.
        - ``"ansi"``: Return a string (``str``) or ``StringIO.StringIO`` containing a
          terminal-style text representation for each timestep. The text can include
          newlines and ANSI escape sequences (e.g. for colors).
        - ``"rgb_array_dict"`` and ``"ansi_dict"``: Return ``dict`` mapping agent ID
          to render frame (RGB or ANSI depending on render mode). Each render frame is
          represents the agent-centric view for the given agent. May also return a
          render for the entire environment (like `"rgb_array"` and `"ansi"` render
          modes) which should be mapped to the `"env"` key in the dictionary by default.

        Note
        ----
        Make sure that your class's :attr:`metadata` ``"render_modes"`` key includes
        the list of supported modes.

        """
        raise NotImplementedError

    def close(self):
        """Close environment and perform any necessary cleanup.

        Should be overridden in subclasses as necessary.
        """
        pass

    @property
    @abc.abstractmethod
    def state(self) -> StateType:
        """The current state for this environment.

        This must be implemented in custom environments.

        Returns
        -------
        StateType

        """

    @property
    def possible_agents(self) -> Tuple[str, ...]:
        """The list of all possible agents that may appear in the environment.

        Returns
        -------
        Tuple[str, ...]

        """
        return self.model.possible_agents

    @property
    def agents(self) -> List[str]:
        """The list of agents active in the environment for current state.

        This will be :attr:`possible_agents`, independent of state, for any environment
        where the number of agents remains constant during and across episodes.

        Returns
        -------
        List[str]

        """
        return self.model.get_agents(self.state)

    @property
    def action_spaces(self) -> Dict[str, spaces.Space]:
        """A mapping from Agent ID to the space of valid actions for that agent.

        Returns
        -------
        Dict[str, spaces.Space]

        """
        return self.model.action_spaces

    @property
    def observation_spaces(self) -> Dict[str, spaces.Space]:
        """A mapping from Agent ID to the space of valid observations for that agent.

        Returns
        -------
        Dict[str, spaces.Space]

        """
        return self.model.observation_spaces

    @property
    def reward_ranges(self) -> Dict[str, Tuple[float, float]]:
        r"""A mapping from Agent ID to the min and max possible rewards for that agent.

        Each reward tuple corresponding to the minimum and maximum possible rewards for
        a given agent for a single step. The default reward range for each agent is set
        to :math:`(-\infty,+\infty)`.

        Returns
        -------
        Dict[str, Tuple[float, float]]

        """
        return self.model.reward_ranges

    @property
    def is_symmetric(self) -> bool:
        """Whether the environment is symmetric.

        An environment is "symmetric" if the ID of an agent in the environment does not
        affect the agent in anyway (i.e. all agents have the same action and observation
        spaces, same reward functions, and there are no differences in initial
        conditions all things considered). Classic examples include Rock-Paper-Scissors,
        Chess, Poker. In "symmetric" environments the same "policy" should do equally
        well independent of the ID of the agent the policy is used for.

        If an environment is not "symmetric" then it is "asymmetric", meaning that
        there are differences in agent properties based on the agent's ID. In
        "asymmetric" environments there is no guarantee that the same "policy" will
        work for different agent IDs. Examples include Pursuit-Evasion games, any
        environments where action and/or observation space differs by agent ID.

        Returns
        -------
        bool
          ``True`` if environment is symmetric, ``False`` if environment is asymmetric.

        """
        return self.model.is_symmetric

    @property
    def unwrapped(self) -> "Env":
        """Completely unwrap this env.

        Returns
        -------
        posggym.Env
            The base non-wrapped posggym.Env instance

        """
        return self

    def __str__(self):
        """Returns a string of the environment with ID of :attr:`spec` if :attr:`spec.

        Returns
        -------
        str
            A string identifying the environment.

        """
        if self.spec is None:
            return f"<{type(self).__name__} instance>"
        else:
            return f"<{type(self).__name__}<{self.spec.id}>>"

    def __enter__(self):
        """Support with-statement for the environment."""
        return self

    def __exit__(self, *args):
        """Support with-statement for the environment."""
        self.close()
        # propagate exception
        return False


class DefaultEnv(Env[StateType, ObsType, ActType]):
    """A default environment implementation using an environment model.

    This class implements the main environment methods - :meth:`reset`,
    :meth:`step`, :attr:`state` - using the environment model.

    Users need only initialize the class with a :class:`posggym.POSGModel` instance by
    calling ``super().__init__(custom_model)`` in their custom environment class that
    inherits from this class.

    The custom environment needs only (optionally) implement rendering and clean-up
    methods:

    - :meth:`render`
    - :meth:`close`

    """

    def __init__(self, model: POSGModel, render_mode: Optional[str] = None):
        assert render_mode is None or render_mode in self.metadata["render_modes"]
        self.model = model
        self.render_mode = render_mode

        self._state = self.model.sample_initial_state()
        self._last_obs = self.model.sample_initial_obs(self._state)
        self._step_num = 0
        self._last_actions: Dict[str, ActType] | None = None
        self._last_rewards: Dict[str, float] | None = None

    def step(
        self, actions: Dict[str, ActType]
    ) -> Tuple[
        Dict[str, ObsType],
        Dict[str, float],
        Dict[str, bool],
        Dict[str, bool],
        bool,
        Dict[str, Dict],
    ]:
        step = self.model.step(self._state, actions)
        self._step_num += 1
        self._state = step.state
        self._last_obs = step.observations
        self._last_actions = actions
        self._last_rewards = step.rewards
        return (
            step.observations,
            step.rewards,
            step.terminations,
            step.truncations,
            step.all_done,
            step.infos,
        )

    def reset(
        self, *, seed: int | None = None, options: Dict[str, Any] | None = None
    ) -> Tuple[Dict[str, ObsType], Dict[str, Dict]]:
        super().reset(seed=seed)
        self._state = self.model.sample_initial_state()
        self._last_obs = self.model.sample_initial_obs(self._state)
        self._last_actions = None
        self._last_rewards = None
        self._step_num = 0
        return self._last_obs, {i: {} for i in self._last_obs}

    @property
    def state(self) -> StateType:
        return copy.copy(self._state)


WrapperStateType = TypeVar("WrapperStateType")
WrapperObsType = TypeVar("WrapperObsType")
WrapperActType = TypeVar("WrapperActType")


class Wrapper(Env[WrapperStateType, WrapperObsType, WrapperActType]):
    """Wraps a :class:`posggym.Env` to allow a modular transformation.

    This class is the base class for all wrappers. Wrappers that inherit from this class
    can modify :attr:`action_spaces`, :attr:`observation_spaces`, :attr:`reward_ranges`
    and :attr:`metadata` attributes , without changing the underlying environment's
    attributes.

    Moreover, the behavior of the :meth:`step` and :meth:`reset` methods can be changed
    by these wrappers. Some attributes (:attr:`spec`, :attr:`render_mode`) will point
    back to the wrapper's environment (i.e. to the corresponding attributes of
    :attr:`env`).

    Note
    ----
    If you inherit from :class:`Wrapper`, don't forget to call ``super().__init__(env)``
    if the subclass overrides the `__init__` method.

    """

    def __init__(self, env: Env[StateType, ObsType, ActType]):
        self.env = env
        self._action_spaces: Dict[str, spaces.Space] | None = None
        self._observation_spaces: Dict[str, spaces.Space] | None = None
        self._reward_ranges: Dict[str, Tuple[float, float]] | None = None
        self._metadata: Dict[str, Any] | None = None

    def __getattr__(self, name):
        """Returns attribute with ``name``, unless ``name`` starts with underscore."""
        if name.startswith("_"):
            raise AttributeError(f"attempted to get missing private attribute '{name}'")
        return getattr(self.env, name)

    @classmethod
    def class_name(cls):
        """Returns the class name of the wrapper."""
        return cls.__name__

    @property
    def model(self) -> POSGModel:
        """Returns the :attr:`Env` :attr:`model`."""
        return self.env.model

    @model.setter
    def model(self, value: POSGModel):
        self.env.model = value

    @property
    def state(self) -> WrapperStateType:
        """Returns the :attr:`Env` :attr:`state`."""
        return self.env.state  # type: ignore

    @property
    def possible_agents(self) -> Tuple[str, ...]:
        """Returns the :attr:`Env` :attr:`possible_agents`."""
        return self.env.possible_agents

    @property
    def agents(self) -> List[str]:
        """Returns the :attr:`Env` :attr:`agents`."""
        return self.env.agents

    @property
    def action_spaces(self) -> Dict[str, spaces.Space]:
        """Return the :attr:`Env` :attr:`action_spaces`.

        This is the :attr:`Env` :attr:`action_spaces` unless it's overwritten then the
        wrapper :attr:`action_spaces` is used.
        """
        if self._action_spaces is None:
            return self.env.action_spaces
        return self._action_spaces

    @action_spaces.setter
    def action_spaces(self, action_spaces: Dict[str, spaces.Space]):
        self._action_spaces = action_spaces

    @property
    def observation_spaces(self) -> Dict[str, spaces.Space]:
        """Return the :attr:`Env` :attr:`observation_spaces`.

        This is the :attr:`Env` :attr:`observation_spaces` unless it's overwritten then
        the wrapper :attr:`observation_spaces` is used.
        """
        if self._observation_spaces is None:
            return self.env.observation_spaces
        return self._observation_spaces

    @observation_spaces.setter
    def observation_spaces(self, observation_spaces: Dict[str, spaces.Space]):
        self._observation_spaces = observation_spaces

    @property
    def reward_ranges(self) -> Dict[str, Tuple[float, float]]:
        """Return the :attr:`Env` :attr:`reward_ranges`.

        This is the :attr:`Env` :attr:`reward_ranges`, unless it's overwritten, then
        the wrapper :attr:`reward_ranges` is used.
        """
        if self._reward_ranges is None:
            return self.env.reward_ranges
        return self._reward_ranges

    @reward_ranges.setter
    def reward_ranges(self, reward_ranges: Dict[str, Tuple[float, float]]):
        self._reward_ranges = reward_ranges

    @property
    def metadata(self) -> Dict[str, Any]:
        """Returns the :attr:`Env` :attr:`metadata`."""
        if self._metadata is None:
            return self.env.metadata
        return self._metadata

    @metadata.setter
    def metadata(self, value: Dict[str, Any]):
        self._metadata = value

    @property
    def spec(self) -> EnvSpec | None:
        """Return the :attr:`Env` :attr:`spec` attribute."""
        return self.env.spec

    @spec.setter
    def spec(self, env_spec: EnvSpec):
        self.env.spec = env_spec

    @property
    def render_mode(self) -> str | None:
        """Return the :attr:`Env` :attr:`render_mode`."""
        return self.env.render_mode

    @render_mode.setter
    def render_mode(self, render_mode: str | None):
        self.env.render_mode = render_mode

    def step(
        self, actions: Dict[str, WrapperActType]
    ) -> Tuple[
        Dict[str, WrapperObsType],
        Dict[str, float],
        Dict[str, bool],
        Dict[str, bool],
        bool,
        Dict[str, Dict],
    ]:
        """Uses the :meth:`step` of the :attr:`env`.

        Can be overwritten to change the returned data.
        """
        return self.env.step(actions)  # type: ignore

    def reset(
        self, *, seed: int | None = None, options: Dict[str, Any] | None = None
    ) -> Tuple[Dict[str, WrapperObsType], Dict[str, Dict]]:
        """Uses the :meth:`reset` of the :attr:`env`.

        Can be overwritten to change the returned data.
        """
        return self.env.reset(seed=seed, options=options)  # type: ignore

    def render(
        self,
    ) -> None | np.ndarray | str | Dict[str, np.ndarray] | Dict[str, str]:
        """Uses the :meth:`render` of the :attr:`env`.

        Can be overwritten to change the returned data.
        """
        return self.env.render()

    def close(self):
        """Closes the wrapper and :attr:`env`."""
        return self.env.close()

    @property
    def unwrapped(self) -> Env:
        """Returns the base environment of the wrapper.

        This will be the bare :class:`posggym.Env` environment, underneath all layers
        of wrappers.
        """
        return self.env.unwrapped

    def __str__(self):
        """Returns the wrapper name and the :attr:`env` representation string."""
        return f"<{type(self).__name__}{self.env}>"

    def __repr__(self):
        """Returns the string representation of the wrapper."""
        return str(self)


class ObservationWrapper(Wrapper[StateType, WrapperObsType, ActType]):
    """Wraps environment to allow modular transformations of observations.

    Subclasses should at least implement the observations function.
    """

    def __init__(self, env: Env[StateType, ObsType, ActType]):
        super().__init__(env)

    def reset(
        self, *, seed: int | None = None, options: Dict[str, Any] | None = None
    ) -> Tuple[Dict[str, WrapperObsType], Dict[str, Dict]]:
        obs, info = self.env.reset(seed=seed, options=options)
        if obs is None:
            return obs, info
        return self.observations(obs), info

    def step(
        self, actions: Dict[str, ActType]
    ) -> Tuple[
        Dict[str, WrapperObsType],
        Dict[str, float],
        Dict[str, bool],
        Dict[str, bool],
        bool,
        Dict[str, Dict],
    ]:
        obs, reward, term, trunc, done, infos = self.env.step(actions)  # type: ignore
        return self.observations(obs), reward, term, trunc, done, infos

    def observations(self, obs: Dict[str, ObsType]) -> Dict[str, WrapperObsType]:
        """Transforms observations received from wrapped environment."""
        raise NotImplementedError


class RewardWrapper(Wrapper[StateType, ObsType, ActType]):
    """Wraps environment to allow modular transformations of rewards.

    Subclasses should at least implement the rewards function.
    """

    def __init__(self, env: Env[StateType, ObsType, ActType]):
        super().__init__(env)

    def step(
        self, actions: Dict[str, ActType]
    ) -> Tuple[
        Dict[str, ObsType],
        Dict[str, float],
        Dict[str, bool],
        Dict[str, bool],
        bool,
        Dict[str, Dict],
    ]:
        obs, reward, term, trunc, done, info = self.env.step(actions)  # type: ignore
        return obs, self.rewards(reward), term, trunc, done, info  # type: ignore

    def rewards(self, rewards: Dict[str, float]) -> Dict[str, float]:
        """Transforms rewards received from wrapped environment."""
        raise NotImplementedError


class ActionWrapper(Wrapper[StateType, ObsType, WrapperActType]):
    """Wraps environment to allow modular transformations of actions.

    Subclasses should at least implement the actions function.
    """

    def __init__(self, env: Env[StateType, ObsType, ActType]):
        super().__init__(env)

    def step(
        self, actions: Dict[str, ActType]
    ) -> Tuple[
        Dict[str, ObsType],
        Dict[str, float],
        Dict[str, bool],
        Dict[str, bool],
        bool,
        Dict[str, Dict],
    ]:
        return self.env.step(self.actions(actions))  # type: ignore

    def actions(self, actions: Dict[str, ActType]) -> Dict[str, WrapperActType]:
        """Transform actions for wrapped environment."""
        raise NotImplementedError



"""Wrapper for recording videos of an environment.

Ref
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/wrappers/record_video.py

"""
import os
from typing import Callable, Optional

from posggym import logger
from posggym.core import Env, Wrapper
from posggym.wrappers.monitoring.video_recorder import VideoRecorder


def capped_cubic_video_schedule(episode_id: int) -> bool:
    """Get cubic schedule.

    Triggers recording of episodes numbers that are perfect cubes :math:`k^3` up to
    the 1000th episode, then every 1000 episodes after that:
    0, 1, 8, 27, 64, 125, 216, 343, 512, 729, 1000, 2000, 3000, ...

    Arguments
    ---------
    episode_id: int
      The episode number

    Returns
    -------
    bool
      Whether to record episode or not.

    """
    if episode_id < 1000:
        return int(round(episode_id ** (1.0 / 3))) ** 3 == episode_id
    else:
        return episode_id % 1000 == 0


class RecordVideo(Wrapper):
    """Wrapper for recording videos of rollouts.

    Arguments
    ---------
    env: posggym.Env
        The environment that will be wrapped
    video_folder: str
        The folder where the recordings will be stored
    episode_trigger: Optional[Callable[[int], bool]]
        Function that accepts an integer and returns ``True`` iff a recording should be
        started at this episode
    step_trigger: Optional[Callable[[int], bool]]
        Function that accepts an integer and returns ``True`` iff a recording should be
        started at this step
    video_length: int
        The length of recorded episodes. If 0, entire episodes are recorded. Otherwise,
        snippets of the specified length are captured
    name_prefix: str, optional
        Will be prepended to the filename of the recordings
    disable_logger: bool
        Whether to disable moviepy logger or not.

    Note
    ----
    This implementation is based on the gymnasium.wrappers.RecordVideo (version
    gymnasium 0.27) wrapper, adapted here to work with posggym's multiagent environment:
    https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/wrappers/record_video.py

    """

    def __init__(
        self,
        env: Env,
        video_folder: str,
        episode_trigger: Optional[Callable[[int], bool]] = None,
        step_trigger: Optional[Callable[[int], bool]] = None,
        video_length: int = 0,
        name_prefix: str = "posggym-video",
        disable_logger: bool = False,
    ):
        super().__init__(env)

        if episode_trigger is None and step_trigger is None:
            episode_trigger = capped_cubic_video_schedule

        trigger_count = sum(x is not None for x in [episode_trigger, step_trigger])
        assert trigger_count == 1, "Must specify exactly one trigger"

        self.episode_trigger = episode_trigger
        self.step_trigger = step_trigger
        self.video_recorder: Optional[VideoRecorder] = None
        self.disable_logger = disable_logger

        self.video_folder = os.path.abspath(video_folder)
        # Create output folder if needed
        if os.path.isdir(self.video_folder):
            logger.warn(
                f"Overwriting existing videos at {self.video_folder} folder (try "
                "specifying a different `video_folder` for the `RecordVideo` wrapper "
                "if this is not desired)"
            )
        os.makedirs(self.video_folder, exist_ok=True)

        self.name_prefix = name_prefix
        self.video_length = video_length
        self.step_id = 0
        self.episode_id = 0

        self.recording = False
        self.episode_done = False
        self.recorded_frames = 0
        self.is_vector_env = getattr(env, "is_vector_env", False)

    def reset(self, **kwargs):
        observations = super().reset(**kwargs)
        self.episode_done = False
        if self.recording:
            assert self.video_recorder is not None
            self.video_recorder.recorded_frames = []
            self.video_recorder.capture_frame()
            self.recorded_frames += 1
            if self.video_length > 0 and self.recorded_frames > self.video_length:
                self.close_video_recorder()
        elif self._video_enabled():
            self.start_video_recorder()
        return observations

    def step(self, actions):
        obs, rewards, terminated, truncated, all_done, info = super().step(actions)

        if not self.episode_done:
            self.episode_done = (not self.is_vector_env and all_done) or (
                self.is_vector_env and all_done[0]
            )

            # increment steps and episodes
            self.step_id += 1
            if self.episode_done:
                self.episode_id += 1

            if self.recording:
                assert self.video_recorder is not None
                self.video_recorder.capture_frame()
                self.recorded_frames += 1
                if self.episode_done or (
                    self.video_length > 0 and self.recorded_frames > self.video_length
                ):
                    self.close_video_recorder()
            elif self._video_enabled():
                self.start_video_recorder()

        return obs, rewards, terminated, truncated, all_done, info

    def render(self, *args, **kwargs):
        """Compute the render frames as specified by environment's render_mode."""
        if self.video_recorder is None or not self.video_recorder.enabled:
            return super().render(*args, **kwargs)

        if len(self.video_recorder.render_history) > 0:
            recorded_frames = [
                self.video_recorder.render_history.pop()
                for _ in range(len(self.video_recorder.render_history))
            ]
            if self.recording:
                return recorded_frames
            else:
                render_output = super().render(*args, **kwargs)
                if isinstance(render_output, dict):
                    if "env" not in render_output:
                        return recorded_frames
                    return recorded_frames + render_output.get("env", [])
                return recorded_frames + render_output
        else:
            return super().render(*args, **kwargs)

    def _video_enabled(self):
        if self.step_trigger:
            return self.step_trigger(self.step_id)
        assert self.episode_trigger is not None
        return self.episode_trigger(self.episode_id)

    def start_video_recorder(self):
        """Starts video recorder."""
        self.close_video_recorder()

        video_name = f"{self.name_prefix}-step-{self.step_id}"
        if self.episode_trigger:
            video_name = f"{self.name_prefix}-episode-{self.episode_id}"

        base_path = os.path.join(self.video_folder, video_name)
        self.video_recorder = VideoRecorder(
            env=self.env,
            base_path=base_path,
            metadata={"step_id": self.step_id, "episode_id": self.episode_id},
            disable_logger=self.disable_logger,
        )

        self.video_recorder.capture_frame()
        self.recorded_frames = 1
        self.recording = True

    def close_video_recorder(self):
        """Closes the video recorder if currently recording."""
        if self.recording:
            assert self.video_recorder is not None
            self.video_recorder.close()
        self.recording = False
        self.recorded_frames = 1

    def close(self):
        super().close()
        self.close_video_recorder()

    def __del__(self):
        self.close_video_recorder()



"""Input processors."""
import abc
from typing import Any

import numpy as np
from gymnasium import spaces


class Processor(abc.ABC):
    """Abstract class for processing inputs.

    Transforms input and input space from one form into another. Inputs can be, for
    example observations or actions.
    """

    def __init__(self, input_space: spaces.Space):
        self.input_space = input_space

    @abc.abstractmethod
    def __call__(self, input: Any) -> Any:
        """Process the input."""

    @abc.abstractmethod
    def unprocess(self, processed_input: Any) -> Any:
        """Unprocess processed input."""

    @abc.abstractmethod
    def get_processed_space(self) -> spaces.Space:
        """Get the processed space."""


class IdentityProcessor(Processor):
    """Identity processor.

    Leaves the input unchanged.
    """

    def __call__(self, input: Any) -> Any:
        return input

    def unprocess(self, processed_input: Any) -> Any:
        return processed_input

    def get_processed_space(self) -> spaces.Space:
        return self.input_space


class FlattenProcessor(Processor):
    """Processor for flattening inputs."""

    def __call__(self, input: Any) -> Any:
        return spaces.flatten(self.input_space, input)

    def unprocess(self, processed_input: Any) -> Any:
        return spaces.unflatten(self.input_space, processed_input)

    def get_processed_space(self) -> spaces.Space:
        return spaces.flatten_space(self.input_space)


class RescaleProcessor(Processor):
    """Rescales input into given range."""

    def __init__(
        self,
        input_space: spaces.Space,
        min_val: float = -1.0,
        max_val: float = 1.0,
        clip: bool = True,
    ):
        assert isinstance(input_space, spaces.Box)
        super().__init__(input_space)
        self.min_val = min_val
        self.max_val = max_val
        self.clip = clip
        self.rescale_factor = (self.max_val - self.min_val) / (
            self.input_space.high - self.input_space.low
        )

    def __call__(self, input: Any) -> Any:
        if self.clip:
            input = np.clip(input, self.input_space.low, self.input_space.high)
        low = self.input_space.low
        return (input - low) * self.rescale_factor + self.min_val

    def unprocess(self, processed_input: Any) -> Any:
        unprocessed_input = (
            processed_input - self.min_val
        ) / self.rescale_factor + self.input_space.low
        if self.clip:
            unprocessed_input = np.clip(
                unprocessed_input, self.input_space.low, self.input_space.high
            )
        return np.array(unprocessed_input, dtype=self.input_space.dtype)

    def get_processed_space(self) -> spaces.Space:
        return spaces.Box(
            low=self.min_val,
            high=self.max_val,
            shape=self.input_space.shape,
            dtype=self.input_space.dtype,
        )



"""Policies for the LevelBasedForaging-v2 environment."""
import os.path as osp

from posggym.agents.grid_world.level_based_foraging import heuristic
from posggym.agents.registration import PolicySpec
from posggym.agents.torch_policy import PPOPolicy
from posggym.agents.utils import processors
from posggym.config import AGENT_MODEL_DIR


agent_model_dir = osp.join(AGENT_MODEL_DIR, "grid_world", "level_based_foraging")

policy_specs = {}
for policy_class in [
    heuristic.LBFHeuristic1,
    heuristic.LBFHeuristic2,
    heuristic.LBFHeuristic3,
    heuristic.LBFHeuristic4,
    heuristic.LBFHeuristic5,
]:
    num = policy_class.__name__.replace("LBFHeuristic", "")
    policy_spec = PolicySpec(
        policy_name=f"H{num}",
        entry_point=policy_class,
        env_id="LevelBasedForaging-v3",
        env_args=None,
        version=0,
        valid_agent_ids=None,
        nondeterministic=False,
        description=policy_class.__doc__.replace("\n", " "),
    )
    policy_specs[policy_spec.id] = policy_spec


for policy_file_name in [
    "RL1.pkl",
    "RL2.pkl",
    "RL3.pkl",
    "RL4.pkl",
    "RL5.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id="LevelBasedForaging-v3",
        env_args={
            "num_agents": 2,
            "max_agent_level": 3,
            "size": 10,
            "max_food": 8,
            "sight": 2,
            "force_coop": False,
            "static_layout": False,
            "observation_mode": "tuple",
        },
        env_args_id="num_agents=2-size=10-static_layout=False",
        policy_file_path=osp.join(
            agent_model_dir,
            "num_agents=2-size=10-static_layout=False",
            policy_file_name,
        ),
        version=0,
        valid_agent_ids=None,
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
        description="Deep RL policy trained using PPO and self-play.",
    )
    policy_specs[spec.id] = spec


for policy_file_name in [
    "RL1.pkl",
    "RL2.pkl",
    "RL3.pkl",
    "RL4.pkl",
    "RL5.pkl",
    "RL6.pkl",
    "RL7.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id="LevelBasedForaging-v3",
        env_args={
            "num_agents": 2,
            "max_agent_level": 3,
            "size": 10,
            "max_food": 8,
            "sight": 2,
            "force_coop": False,
            "static_layout": True,
            "observation_mode": "tuple",
        },
        env_args_id="num_agents=2-size=10-static_layout=True",
        policy_file_path=osp.join(
            agent_model_dir,
            "num_agents=2-size=10-static_layout=True",
            policy_file_name,
        ),
        version=0,
        valid_agent_ids=None,
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
        description="Deep RL policy trained using PPO and self-play.",
    )
    policy_specs[spec.id] = spec



"""Heuristic policies for Predator-Prey continuous environment."""
from __future__ import annotations

import abc
import math
from typing import TYPE_CHECKING, Tuple, cast

import numpy as np

from posggym.agents.policy import Policy, PolicyID, PolicyState
from posggym.agents.utils import action_distributions
from posggym.envs.continuous.predator_prey_continuous import (
    PPAction,
    PPObs,
    PredatorPreyContinuousModel,
)
from posggym.utils import seeding

if TYPE_CHECKING:
    from posggym.posggym.model import POSGModel
    from posggym.utils.history import AgentHistory


class PPCHeuristicPolicy(Policy[PPAction, PPObs], abc.ABC):
    """Base class for heuristic policies for Predator-Prey continuous environment."""

    def __init__(self, model: POSGModel, agent_id: str, policy_id: PolicyID):
        super().__init__(model, agent_id, policy_id)
        self.model = cast(PredatorPreyContinuousModel, model)
        self._rng, _ = seeding.np_random()

        self.action_space = self.model.action_spaces[self.agent_id]
        self.action_dtype = self.action_space.dtype
        self.obs_dist = self.model.obs_dist
        self.n_sensors = self.model.n_sensors
        self.sensor_angle = 2 * math.pi / self.n_sensors

    def reset(self, *, seed: int | None = None):
        super().reset(seed=seed)
        if seed is not None:
            self._rng, _ = seeding.np_random(seed=seed)

    def get_initial_state(self) -> PolicyState:
        state = super().get_initial_state()
        state.update(
            {
                "pi": None,
                "last_obs": None,
            }
        )
        return state

    def get_next_state(
        self,
        action: PPAction | None,
        obs: PPObs,
        state: PolicyState,
    ) -> PolicyState:
        pi = self._get_pi_from_obs(obs)
        return {
            "pi": pi,
            "last_obs": obs,
        }

    def get_state_from_history(self, history: AgentHistory) -> PolicyState:
        _, last_obs = history.get_last_step()
        if last_obs is not None:
            return {
                "pi": self._get_pi_from_obs(last_obs),
                "last_obs": last_obs,
            }
        return self.get_initial_state()

    def sample_action(self, state: PolicyState) -> PPAction:
        if state["pi"] is None:
            raise ValueError(
                "Policy state does not contain a valid action distribution. Make sure"
                "to call `step` or `get_next_state` before calling `sample_action`"
            )
        return np.array(state["pi"].sample(), dtype=self.action_dtype)

    def get_pi(self, state: PolicyState) -> action_distributions.ActionDistribution:
        return state["pi"]

    def get_value(self, state: PolicyState) -> float:
        raise NotImplementedError(
            f"`get_value()` no implemented by {self.__class__.__name__} policy"
        )

    @abc.abstractmethod
    def _get_pi_from_obs(self, obs: PPObs) -> action_distributions.ActionDistribution:
        raise NotImplementedError

    def _get_closest_prey(self, obs: PPObs) -> Tuple[float, float] | None:
        prey_obs = obs[2 * self.n_sensors : 3 * self.n_sensors]
        closest_idx = np.argmin(prey_obs)
        if prey_obs[closest_idx] == self.model.obs_dist:
            return None
        closest_angle = closest_idx * self.sensor_angle
        closest_dist = prey_obs[closest_idx]
        return closest_dist, closest_angle

    def _get_closest_predator(self, obs: PPObs) -> Tuple[float, float] | None:
        pred_obs = obs[self.n_sensors : 2 * self.n_sensors]
        closest_idx = np.argmin(pred_obs)
        if pred_obs[closest_idx] == self.model.obs_dist:
            return None
        closest_angle = closest_idx * self.sensor_angle
        closest_dist = pred_obs[closest_idx]
        return closest_dist, closest_angle

    def _get_closest_prey_to_predator(
        self, obs: PPObs, pred_dist: float, pred_angle: float
    ) -> Tuple[float, float] | None:
        # find prey with minimum distance to predator
        # d^2 = P^2 + p^2 - 2Pp cos(theta)
        # d = distance between predator and prey
        # P = distance to closest predator
        # p = distance to prey
        # theta = angle between predator and prey (from ego's perspective)
        # skip sqrt since we just want to find the closest prey idx, not actual dist
        prey_obs = obs[2 * self.n_sensors : 3 * self.n_sensors]
        prey_dists = (
            pred_dist**2
            + prey_obs**2
            - (
                2
                * pred_dist
                * prey_obs
                * np.cos(pred_angle - np.arange(self.n_sensors) * self.sensor_angle)
            )
        )
        prey_dists = np.where(prey_obs < self.model.obs_dist, prey_dists, np.inf)
        if np.all(prey_dists == np.inf):
            return None
        closest_prey_idx = np.argmin(prey_dists)
        closest_prey_angle = closest_prey_idx * self.sensor_angle
        closest_prey_dist = prey_dists[closest_prey_idx]
        return closest_prey_dist, closest_prey_angle

    def _get_action_to_target(
        self, target_dist: float, target_angle: float
    ) -> PPAction:
        dyaw = self._get_angular_velocity(target_angle)
        dvel = self._get_linear_velocity(target_dist)
        return np.array([dyaw, dvel], dtype=np.float32)

    def _get_angular_velocity(self, target_angle: float) -> float:
        if target_angle > math.pi:
            dyaw = max(target_angle - (2 * math.pi), -self.model.dyaw_limit)
        else:
            dyaw = min(target_angle, self.model.dyaw_limit)
        return dyaw

    def _get_linear_velocity(self, target_dist: float) -> float:
        return min(target_dist, 1.0)


class PPCHeuristic0Policy(PPCHeuristicPolicy):
    """Heuristic 0 policy for Predator-Prey continuous environment.

    Predator greedily moves towards closest observed prey, or moves randomly.
    """

    def _get_pi_from_obs(self, obs: PPObs) -> action_distributions.ActionDistribution:
        target = self._get_closest_prey(obs)
        if target is None:
            # No prey or predator in sight
            pi = action_distributions.ContinousUniformActionDistribution(
                low=self.action_space.low, high=self.action_space.high, rng=self._rng
            )
        else:
            # Move towards closest prey
            target_dist, target_angle = target
            action = self._get_action_to_target(target_dist, target_angle)
            pi = action_distributions.DeterministicActionDistribution(action)

        return pi


class PPCHeuristic1Policy(PPCHeuristicPolicy):
    """Heuristic 1 policy for Predator-Prey continuous environment.

    Predator greedily moves towards closest observed prey, closest observed predator,
    or moves randomly, in that order.
    """

    def _get_pi_from_obs(self, obs: PPObs) -> action_distributions.ActionDistribution:
        target = self._get_closest_prey(obs)
        if target is None:
            target = self._get_closest_predator(obs)

        if target is None:
            # No prey or predator in sight
            pi = action_distributions.ContinousUniformActionDistribution(
                low=self.action_space.low, high=self.action_space.high, rng=self._rng
            )
        else:
            # Move towards closest prey or predator
            target_dist, target_angle = target
            action = self._get_action_to_target(target_dist, target_angle)
            pi = action_distributions.DeterministicActionDistribution(action)

        return pi


class PPCHeuristic2Policy(PPCHeuristicPolicy):
    """Heuristic 2 policy for Predator-Prey continuous environment.

    Predator moves randomly until it observes prey, then it greedily moves towards
    closest observed prey, or if it observes a prey and predator then it moves
    towards the closest observed prey to the closest observed predator.

    """

    def _get_pi_from_obs(self, obs: PPObs) -> action_distributions.ActionDistribution:
        prey = self._get_closest_prey(obs)
        if prey is None:
            # No prey in sight
            return action_distributions.ContinousUniformActionDistribution(
                low=self.action_space.low, high=self.action_space.high, rng=self._rng
            )

        pred = self._get_closest_predator(obs)
        if pred is None:
            # No predator in sight, so move towards closest prey
            action = self._get_action_to_target(prey[0], prey[1])
            return action_distributions.DeterministicActionDistribution(action)
        pred_dist, pred_angle = pred

        # Move towards closest prey to predator
        prey = self._get_closest_prey_to_predator(obs, pred_dist, pred_angle)
        assert prey is not None
        action = self._get_action_to_target(prey[0], prey[1])
        return action_distributions.DeterministicActionDistribution(action)


class PPCHeuristic3Policy(PPCHeuristicPolicy):
    """Heuristic 3 policy for Predator-Prey continuous environment.

    Predator moves randomly until it observes a predator, then it greedily moves towards
    closest observed predator, or if it observes a prey and predator then it moves
    towards the closest observed prey to the closest observed predator.

    """

    def _get_pi_from_obs(self, obs: PPObs) -> action_distributions.ActionDistribution:
        pred = self._get_closest_predator(obs)
        if pred is None:
            # No predator in sight
            return action_distributions.ContinousUniformActionDistribution(
                low=self.action_space.low, high=self.action_space.high, rng=self._rng
            )
        pred_dist, pred_angle = pred

        prey = self._get_closest_prey_to_predator(obs, pred_dist, pred_angle)
        if prey is None:
            # No prey in sight, so move towards predator
            action = self._get_action_to_target(pred_dist, pred_angle)
            return action_distributions.DeterministicActionDistribution(action)

        # Move towards closest prey to predator
        action = self._get_action_to_target(prey[0], prey[1])
        return action_distributions.DeterministicActionDistribution(action)



"""Shortest path policy for Driving envs."""
from __future__ import annotations

from queue import PriorityQueue
from typing import TYPE_CHECKING, Dict, Set, Tuple, cast

from posggym.agents.policy import Policy, PolicyID, PolicyState
from posggym.agents.utils import action_distributions
from posggym.envs.grid_world.core import Coord, Direction
from posggym.envs.grid_world.driving import (
    ACCELERATE,
    DECELERATE,
    DO_NOTHING,
    INIT_DIR,
    INIT_SPEED,
    TURN_LEFT,
    TURN_RIGHT,
    VEHICLE,
    DAction,
    DObs,
    DrivingModel,
    Speed,
)
from posggym.utils import seeding

if TYPE_CHECKING:
    from posggym.model import POSGModel
    from posggym.utils.history import AgentHistory


# Current coord, speed, facing direction
Pos = Tuple[Coord, Speed, Direction]


class DrivingShortestPathPolicy(Policy[DAction, DObs]):
    """Shortest Path Policy for the Driving environment.

    This policy selects actions that lead it to taking the shortest path to the agent's
    goal. If there are multiple actions on the shortest path then selects uniformly
    at random from those actions.

    Arguments
    ---------
    aggressiveness : float
        The aggressiveness of the policy towards other vehicles. A value of 0.0 means
        the policy will always stop when it sees another vehicle, while a value of 1.0
        means the policy will ignore other vehicles and always take the shortest path
        action. Values in between will change how far away another vehicle needs to
        be before the policy will stop. Furthermore, if `aggressiveness < 0.5` vehicles
        will never accelerate to max speed (Default=`1.0`).

    """

    def __init__(
        self,
        model: POSGModel,
        agent_id: str,
        policy_id: PolicyID,
        aggressiveness: float = 1.0,
        precompute_shortest_paths: bool = False,
    ):
        super().__init__(model, agent_id, policy_id)
        self.model = cast(DrivingModel, model)
        assert (
            0.0 <= aggressiveness <= 1.0
        ), f"Aggressiveness must be between 0.0 and 1.0, got {aggressiveness}"
        self.aggressiveness = aggressiveness
        self._grid = self.model.grid
        self._action_space = list(range(self.model.action_spaces[agent_id].n))
        self._rng, _ = seeding.std_random()

        obs_front, obs_back, obs_side = self.model.obs_dim
        self.obs_width = (2 * obs_side) + 1
        self.agent_obs_idx = (obs_front * self.obs_width) + obs_side
        self.agent_obs_coord = (obs_side, obs_front)
        self.max_obs_dist = max(obs_front, obs_back) + obs_side

        self.shortest_paths = {}
        if precompute_shortest_paths:
            for dest_coord in set.union(*self._grid.dest_coords):
                self.shortest_paths[dest_coord] = self.get_all_shortest_paths(
                    dest_coord
                )

    def reset(self, *, seed: int | None = None):
        super().reset(seed=seed)
        if seed is not None:
            self._rng, _ = seeding.std_random(seed=seed)

    def get_initial_state(self) -> PolicyState:
        state = super().get_initial_state()
        state.update(
            {
                "update_num": 0,
                "speed": INIT_SPEED,
                "facing_dir": INIT_DIR,
                "coord": (0, 0),
                # store previous coord to avoid getting stuck in same position
                "prev_coord": (0, 0),
                "dest_coord": (0, 0),
                "min_other_vehicle_dist": 0,
            }
        )
        return state

    def get_next_state(
        self,
        action: DAction | None,
        obs: DObs,
        state: PolicyState,
    ) -> PolicyState:
        if state["update_num"] == 0:
            # I.e. first observation with no action performed yet
            next_facing_dir = INIT_DIR
        else:
            next_facing_dir = self.model.get_next_direction(
                action, obs[1], state["facing_dir"]
            )

        next_state = {
            "update_num": state["update_num"] + 1,
            "speed": obs[1],
            "facing_dir": next_facing_dir,
            "coord": obs[2],
            "prev_coord": state["coord"],
            "dest_coord": obs[3],
            "min_other_vehicle_dist": self.get_min_other_vehicle_dist(obs[0]),
        }
        return next_state

    def get_state_from_history(self, history: AgentHistory) -> PolicyState:
        _, obs = history.get_last_step()
        assert obs is not None
        facing_dir = INIT_DIR
        for a, _ in history:
            if a is None:
                continue
            facing_dir = self.model.get_next_direction(a, facing_dir)
        return {
            "update_num": len(history),
            "speed": obs[1],
            "facing_dir": facing_dir,
            "coord": obs[2],
            "prev_coord": (0, 0) if len(history) == 1 else history[-2][1][2],
            "dest_coord": obs[3],
            "min_other_vehicle_dist": self.get_min_other_vehicle_dist(obs[0]),
        }

    def sample_action(self, state: PolicyState) -> DAction:
        return self.get_pi(state).sample()

    def get_pi(self, state: PolicyState) -> action_distributions.ActionDistribution:
        vehicle_dist = state["min_other_vehicle_dist"]
        if vehicle_dist <= self.max_obs_dist:
            proximity = (vehicle_dist - 1) / (self.max_obs_dist - 1)
            if proximity < (1 - self.aggressiveness):
                if state["speed"] > Speed.STOPPED:
                    pi = {DECELERATE: 1.0}
                elif state["speed"] < Speed.STOPPED:
                    pi = {ACCELERATE: 1.0}
                else:
                    pi = {DO_NOTHING: 1.0}
                return action_distributions.DiscreteActionDistribution(pi, self._rng)

        # get distances to dest for each action
        dists = []
        for a in self._action_space:
            if (
                self.aggressiveness < 0.5
                and state["speed"] >= Speed.FORWARD_SLOW
                and a == ACCELERATE
            ):
                # don't accelerate if aggressiveness is low and agent is already
                # moving at FORWARD_SLOW or faster
                dists.append(float("inf"))
                continue

            a_speed = self.model.get_next_speed(a, state["speed"])
            a_facing_dir = self.model.get_next_direction(
                a, a_speed, state["facing_dir"]
            )
            a_move_dir = self.model.get_move_direction(a, a_speed, state["facing_dir"])
            next_coord = state["coord"]
            for _ in range(abs(a_speed - Speed.STOPPED)):
                next_coord = self._grid.get_next_coord(
                    next_coord, a_move_dir, ignore_blocks=False
                )

            d = self.get_dest_shortest_path_dist(
                state["dest_coord"], (next_coord, a_speed, a_facing_dir)
            )
            dists.append(d)

        # get pi from dists
        min_dist = min(dists)
        num_min = dists.count(min_dist)
        pi = {}
        for a in self._action_space:
            if dists[a] == min_dist:
                pi[a] = 1.0 / num_min
            else:
                pi[a] = 0.0

        return action_distributions.DiscreteActionDistribution(pi, self._rng)

    def get_value(self, state: PolicyState) -> float:
        raise NotImplementedError(
            f"`get_value()` no implemented by {self.__class__.__name__} policy"
        )

    def get_min_other_vehicle_dist(self, local_obs: Tuple[int, ...]) -> int:
        """Get minimum distance to other vehicle in local observation."""
        min_other_vehicle_dist = self.max_obs_dist + 1
        for idx, cell_obs in enumerate(local_obs):
            if idx != self.agent_obs_idx and cell_obs == VEHICLE:
                col, row = idx % self.obs_width, idx // self.obs_width
                dist = abs(col - self.agent_obs_coord[0]) + abs(
                    row - self.agent_obs_coord[1]
                )
                min_other_vehicle_dist = min(dist, min_other_vehicle_dist)
        return min_other_vehicle_dist

    def get_dest_shortest_path_dist(self, dest_coord: Coord, pos: Pos) -> int:
        """Get shortest path distance to destination coord from given position."""
        if dest_coord not in self.shortest_paths:
            self.shortest_paths[dest_coord] = self.get_all_shortest_paths(dest_coord)

        min_dist = float("inf")
        for dists in self.shortest_paths[dest_coord].values():
            if dists.get(pos, float("inf")) < min_dist:
                min_dist = dists[pos]
        return min_dist

    def get_all_shortest_paths(self, origin: Coord) -> Dict[Pos, Dict[Pos, int]]:
        """Get shortest paths from given origin to all other positions.

        Note, this is a search over vehicle configurations, i.e. (coord, speed,
        facing_dir), rather than just vehicle coordinate. This yields a shortest path
        distance in terms of number of actions, rather than number of cells.
        """
        src_dists = {}
        for speed in Speed:
            for facing_dir in Direction:
                pos = (origin, speed, facing_dir)
                src_dists[pos] = self.dijkstra(pos)
        return src_dists

    def dijkstra(self, origin: Pos) -> Dict[Pos, int]:
        """Get shortest path distance to origin from all other positions."""
        dist = {origin: 0}
        pq = PriorityQueue()  # type: ignore
        pq.put((dist[origin], origin))

        visited = {origin}

        while not pq.empty():
            _, pos = pq.get()

            for adj_pos in self.get_prev_positions(pos):
                if dist[pos] + 1 < dist.get(adj_pos, float("inf")):
                    dist[adj_pos] = dist[pos] + 1

                    if adj_pos not in visited:
                        pq.put((dist[adj_pos], adj_pos))
                        visited.add(adj_pos)
        return dist

    def get_prev_positions(self, pos: Pos) -> Set[Pos]:
        """Get all positions reachable from given position."""
        coord, speed, facing_dir = pos
        prev_positions = set()
        for prev_a in [DO_NOTHING, TURN_LEFT, TURN_RIGHT, ACCELERATE, DECELERATE]:
            if prev_a in [DO_NOTHING, ACCELERATE, DECELERATE] or speed == Speed.REVERSE:
                # agent moved in straight line from prev pos
                if prev_a == DECELERATE:
                    prev_speed = Speed(min(speed + 1, Speed.FORWARD_FAST))
                    prev_dir = facing_dir
                elif prev_a == ACCELERATE:
                    prev_speed = Speed(max(speed - 1, Speed.REVERSE))
                    prev_dir = facing_dir
                else:
                    # DO_NOTHING or (TURN_LEFT or TURN_RIGHT in REVERSE)
                    # Agent cannot turn while reversing, so action is same as DO_NOTHING
                    prev_speed = speed
                    prev_dir = facing_dir

                if speed == Speed.REVERSE:
                    move_dir = prev_dir
                else:
                    move_dir = Direction((prev_dir + 2) % len(Direction))

                next_coord = self._grid.get_next_coord(
                    coord,
                    Direction((move_dir + 2) % len(Direction)),
                    ignore_blocks=False,
                )
                next_coord_blocked = next_coord == coord

                prev_coord = coord
                for _ in range(abs(prev_speed - Speed.STOPPED)):
                    if next_coord_blocked:
                        # if next coord is blocked then may have moved to current
                        # coord from 0, 1 or 2 cells away
                        prev_positions.add((prev_coord, prev_speed, prev_dir))
                    prev_coord = self._grid.get_next_coord(
                        prev_coord, move_dir, ignore_blocks=False
                    )
                prev_positions.add((prev_coord, prev_speed, prev_dir))
            else:
                # TURN_RIGHT or TURN_LEFT (and not in REVERSE)
                # agent turned and maybe moved from prev pos
                if prev_a == TURN_RIGHT:
                    prev_dir = Direction((facing_dir - 1) % len(Direction))
                else:
                    prev_dir = Direction((facing_dir + 1) % len(Direction))

                if speed == Speed.STOPPED:
                    # agent turned in place
                    next_coord = coord
                    prev_speed = speed
                    prev_positions.add((coord, prev_speed, prev_dir))
                elif speed == Speed.FORWARD_FAST:
                    # not possible, since turning reduces speed to FORWARD_SLOW
                    continue
                else:
                    # agent turned and moved
                    # if current speed is FORWARD_SLOW, then agent could have
                    # been moving at FORWARD_FAST or FORWARD_SLOW
                    prev_speeds = [speed]
                    if speed == Speed.FORWARD_SLOW:
                        prev_speeds.append(Speed.FORWARD_FAST)

                    prev_coords = set()
                    # Need to check if next coord is blocked, since agent may have
                    # turned into a wall and so stayed at same coord, but changed dir
                    next_coord = self._grid.get_next_coord(
                        coord, facing_dir, ignore_blocks=False
                    )
                    if next_coord == coord:
                        prev_coords.add(coord)

                    # agent moved from cell in opposite direction to facing dir
                    move_dir = Direction((facing_dir + 2) % len(Direction))
                    prev_coord = self._grid.get_next_coord(
                        coord, move_dir, ignore_blocks=False
                    )
                    prev_coords.add(prev_coord)

                    for prev_coord in prev_coords:
                        for prev_speed in prev_speeds:
                            prev_positions.add((prev_coord, prev_speed, prev_dir))

        return prev_positions



"""Classes for recording videos of episodes.

Ref:
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/wrappers/monitoring/video_recorder.py

"""
import json
import os
import os.path
import tempfile
from typing import Dict, Optional, List

from posggym import Env, error, logger


class VideoRecorder:
    """VideoRecorder renders a nice movie of a rollout, frame by frame.

    It comes with an `enabled` option so you can still use the same code on
    episodes where you don't want to record video.

    This implementation is the same as the gymnasium VideoRecorder class but
    makes it compatible with posggym.env.render function which returns rgb
    arrays for the whole environment as well as (optionally) each agent.

    Note
    ----
    You are responsible for calling `close` on a created VideoRecorder, or else
    you may leak an encoder process.

    """

    combatible_render_modes = ["rgb_array", "rgb_array_dict"]

    def __init__(
        self,
        env: Env,
        path: Optional[str] = None,
        metadata: Optional[Dict] = None,
        enabled: bool = True,
        base_path: Optional[str] = None,
        disable_logger: bool = False,
    ):
        """Video recorder renders a nice movie of a rollout, frame by frame.

        Arguments
        ---------
        env: Environment to take video of.
        path: Path to the video file; will be randomly chosen if omitted.
        metadata: Contents to save to the metadata file.
        enabled: Whether to actually record video, or just no-op (for convenience)
        base_path: Alternatively, path to the video file without extension, which will
            be added.
        disable_logger: Whether to disable moviepy logger or not.

        """
        try:
            # check that moviepy is now installed
            import moviepy  # noqa
        except ImportError as e:
            raise error.DependencyNotInstalled(
                "MoviePy is not installed, run `pip install moviepy`"
            ) from e

        self._async = env.metadata.get("semantics.async")
        self.enabled = enabled
        self.disable_logger = disable_logger
        self._closed = False

        self.render_history: List = []
        self.env = env

        self.render_mode = env.render_mode

        if self.render_mode not in self.combatible_render_modes:
            logger.warn(
                f"Disabling video recorder because environment {env} was not "
                "initialized with any compatible video modes in "
                f"{self.combatible_render_modes}."
            )
            # Disable since the env not initialized with a compatible `render_mode`
            self.enabled = False

        # Don't bother setting anything else if not enabled
        if not self.enabled:
            return

        if path is not None and base_path is not None:
            raise error.Error("You can pass at most one of `path` or `base_path`.")

        self.last_frame = None
        self.env = env

        required_ext = ".mp4"
        if path is None:
            if base_path is not None:
                # Base path given, append ext
                path = base_path + required_ext
            else:
                # Otherwise, just generate a unique filename
                with tempfile.NamedTemporaryFile(suffix=required_ext) as f:
                    path = f.name
        self.path = path

        path_base, actual_ext = os.path.splitext(self.path)

        if actual_ext != required_ext:
            raise error.Error(
                f"Invalid path given: {self.path} -- must have file extension "
                f"{required_ext}."
            )

        self.frames_per_sec = env.metadata.get("render_fps", 30)

        self.broken = False

        # Dump metadata
        self.metadata = metadata or {}
        self.metadata["content_type"] = "video/mp4"
        self.metadata_path = f"{path_base}.meta.json"
        self.write_metadata()

        logger.info("Starting new video recorder writing to %s", self.path)
        self.recorded_frames: List = []

    @property
    def functional(self) -> bool:
        """Returns if the video recorder is functional, is enabled and not broken."""
        return self.enabled and not self.broken

    def capture_frame(self):
        """Render the given `env` and add the resulting frame to the video.

        Note, this function has been changed from the source implementation.
        """
        if not self.functional:
            return

        frame = self.env.render()
        if isinstance(frame, list):
            self.render_history += frame
            frame = frame[-1]
        elif isinstance(frame, dict):
            if "env" not in frame:
                logger.warn(
                    "The video recorder expects an entry with the key `env` when "
                    "trying to record an environment that is using the "
                    "`rgb_array_dict` render mode."
                )
                self.broken = True
                return
            frame = frame["env"]

        if self._closed:
            logger.warn(
                "The video recorder has been closed and no frames will be "
                "captured anymore."
            )
            return
        logger.debug("Capturing video frame: path=%s", self.path)

        if frame is None:
            if self._async:
                return
            else:
                # Indicates a bug in the environment: don't want to raise
                # an error here.
                logger.warn(
                    "Env returned None on `render()`. Disabling further rendering for "
                    f"video recorder by marking as disabled: path={self.path} "
                    f"metadata_path={self.metadata_path}"
                )
                self.broken = True
        else:
            self.recorded_frames.append(frame)

    def close(self):
        """Flush all data to disk and close any open frame encoders."""
        if not self.enabled or self._closed:
            # mark as closed in case it is broken to save future warnings
            self._closed = True
            return

        if len(self.recorded_frames) > 0:
            try:
                from moviepy.video.io.ImageSequenceClip import ImageSequenceClip
            except ImportError as e:
                raise error.DependencyNotInstalled(
                    "MoviePy is not installed, run `pip install moviepy`"
                ) from e

            clip = ImageSequenceClip(self.recorded_frames, fps=self.frames_per_sec)
            moviepy_logger = None if self.disable_logger else "bar"
            clip.write_videofile(self.path, logger=moviepy_logger)
        else:
            # No frames captured. Set metadata.
            if self.metadata is None:
                self.metadata = {}
            self.metadata["empty"] = True

        self.write_metadata()

        # Stop tracking this for autoclose
        self._closed = True

    def write_metadata(self):
        """Writes metadata to metadata path."""
        with open(self.metadata_path, "w") as f:
            json.dump(self.metadata, f)

    def __del__(self):
        """Closes the environment correctly when the recorder is deleted."""
        # Make sure we've closed up shop when garbage collecting
        if not self._closed:
            logger.warn("Unable to save last video! Did you call close()?")



"""Functions for pairwise evaluation of agents."""
from __future__ import annotations

import csv
import multiprocessing as mp
import os
from datetime import datetime
from itertools import product
from typing import Dict, List, NamedTuple, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import torch

import posggym
import posggym.agents as pga
import posggym.model as M
from posggym import logger
from posggym.config import BASE_RESULTS_DIR


POLICIES_TO_SKIP = [
    "DiscreteFixedDistributionPolicy-v0",  # this is the same as Random-v0 by default
]

global lock
lock = None


def _init_pool_processes(write_lock: mp.Lock):
    # ref:
    # https://stackoverflow.com/questions/69907453/lock-objects-should-only-be-shared-between-processes-through-inheritance
    global lock
    lock = write_lock


class PWCParams(NamedTuple):
    """Pairwise Comparison Parameters."""

    env_id: str
    env_args_id: str | None
    env_args: Dict[str, str]
    policy_ids: Dict[str, str]
    output_file: str
    num_episodes: int
    seed: int | None
    run_num: int
    verbose: bool = False


def run_episodes(args) -> Dict[str, Dict[str, float]]:
    """Run episodes and return the average reward for each policy."""
    params, total_runs = args
    policy_names = {
        i: pga.spec(policy_id).policy_name for i, policy_id in params.policy_ids.items()
    }
    if params.verbose:
        print(
            f"- Running episodes for {params.env_id}/{params.env_args_id} with "
            f"policies={policy_names}"
        )

    env = posggym.make(params.env_id, **params.env_args)
    policies = {
        i: pga.make(policy_id, env.model, i)
        for i, policy_id in params.policy_ids.items()
    }

    obs, _ = env.reset(seed=params.seed)
    for idx, policy in enumerate(policies.values()):
        policy.reset(seed=params.seed + idx if params.seed is not None else None)

    episode_rewards = {i: np.zeros(params.num_episodes) for i in policies}
    episode_lens = {i: np.zeros(params.num_episodes) for i in policies}
    episode_outcomes = {i: {outcome: 0 for outcome in M.Outcome} for i in policies}
    for ep_num in range(params.num_episodes):
        obs, _ = env.reset()
        for policy in policies.values():
            policy.reset()

        all_done = False
        while not all_done:
            actions = {}
            for i in env.agents:
                if policies[i].observes_state:
                    actions[i] = policies[i].step(env.state)
                else:
                    actions[i] = policies[i].step(obs[i])
            obs, rewards, _, _, all_done, info = env.step(actions)
            for i, r in rewards.items():
                episode_rewards[i][ep_num] += r
                episode_lens[i][ep_num] += 1

            if all_done:
                for i, info_i in info.items():
                    if "outcome" in info_i:
                        episode_outcomes[i][info_i["outcome"]] += 1

    agent_ids = sorted(params.policy_ids.keys())
    rows = []
    for i in agent_ids:
        policy_id = params.policy_ids[i]
        policy_spec = pga.spec(policy_id)

        co_player_names = [policy_names[j] for j in agent_ids if j != i]
        co_team_id = "(" + ",".join(co_player_names) + ")"

        row = {
            "env_id": params.env_id,
            "env_args_id": params.env_args_id,
            "symmetric": env.is_symmetric,
            "seed": params.seed,
            "num_agents": len(params.policy_ids),
            "num_episodes": params.num_episodes,
            "policy_id": policy_id,
            "policy_name": policy_spec.policy_name,
            "policy_version": policy_spec.version,
            "agent_id": i,
            "co_team_id": co_team_id,
            "episode_reward_mean": np.mean(episode_rewards[i]),
            "episode_reward_std": np.std(episode_rewards[i]),
        }
        for outcome in M.Outcome:
            row[f"num_{outcome.name}"] = episode_outcomes[i][outcome]
        rows.append(row)

    if lock is not None:
        lock.acquire()

    with open(params.output_file, "a", newline="") as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=list(rows[0].keys()))
        for row in rows:
            writer.writerow(row)

    if lock is not None:
        lock.release()

    if params.run_num % max(10, total_runs // 10) == 0:
        print(f"Run {params.run_num}/{total_runs} complete.")


def get_pairwise_comparison_params(
    env_id: str,
    output_dir: str,
    env_args_id: str | None = None,
    num_episodes: int = 1000,
    seed: int | None = None,
    verbose: bool = False,
) -> List[PWCParams]:
    """Get parameters for pairwise comparisons of all of an environment's policies."""
    # attempt to make env to check if it is registered (displays nicer error msg)
    posggym.make(env_id)

    env_output_dir = os.path.join(output_dir, env_id)
    os.makedirs(env_output_dir, exist_ok=True)

    # construct list of all possible policy pairings for the environment
    env_args_map = pga.get_all_envs()[env_id]
    all_pairwise_comparisons = []
    run_num = 1
    for args_id, env_args in env_args_map.items():
        if env_args_id is not None and args_id != env_args_id:
            continue

        if verbose:
            print(f"- Running for {args_id=}")

        env_args = env_args or {}
        env = posggym.make(env_id, **env_args)

        all_agent_policies = pga.get_env_agent_policies(
            env_id, env_args, include_generic_policies=True
        )
        agent_ids = list(all_agent_policies.keys())
        if verbose:
            print(f"  - {len(agent_ids)} agents")

        output_file = os.path.join(env_output_dir, f"{args_id}.csv")
        if os.path.exists(output_file):
            logger.warn(f"{output_file} exists. Rewriting.")

        headers = [
            "env_id",
            "env_args_id",
            "symmetric",
            "seed",
            "num_agents",
            "num_episodes",
            "policy_id",
            "policy_name",
            "policy_version",
            "agent_id",
            "co_team_id",
            "episode_reward_mean",
            "episode_reward_std",
        ]
        for outcome in M.Outcome:
            headers.append(f"num_{outcome.name}")

        with open(output_file, "w", newline="") as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=headers)
            writer.writeheader()

        pairwise_policy_ids = []
        if env.is_symmetric:
            # run one policy against multiple versions of the same policy for all other
            # agents (otherwise the # of policies can be too large for N > 2)
            all_policy_specs = []
            for policy_spec in all_agent_policies[agent_ids[0]]:
                if policy_spec.id not in POLICIES_TO_SKIP:
                    all_policy_specs.append(policy_spec)
                elif verbose:
                    print(f"  - Skipping {policy_spec.id}")

            if verbose:
                print(f"  - {len(all_policy_specs)} policies to compare")

            for i in agent_ids:
                for policy_spec_i, policy_spec_j in product(all_policy_specs, repeat=2):
                    if policy_spec_i.id == policy_spec_j.id and i != agent_ids[0]:
                        # only add same vs same policy once
                        continue
                    policy_ids = {}
                    for j in agent_ids:
                        if i == j:
                            policy_ids[j] = policy_spec_i.id
                        else:
                            policy_ids[j] = policy_spec_j.id
                    pairwise_policy_ids.append(policy_ids)
        else:
            all_policy_specs = []
            num_policies = {}
            for i in agent_ids:
                agent_policy_specs = []
                for policy_spec in all_agent_policies[i]:
                    if policy_spec.id not in POLICIES_TO_SKIP:
                        agent_policy_specs.append(policy_spec)
                    else:
                        print(f"  - Skipping {policy_spec.id}")
                all_policy_specs.append(agent_policy_specs)
                num_policies[i] = len(agent_policy_specs)

            if verbose:
                print(f"  - policies to compare {num_policies}")

            # run all pairwise combinations of policies
            for policy_specs in product(*all_policy_specs):
                policy_ids = {i: spec.id for i, spec in zip(agent_ids, policy_specs)}
                # print(f"  {policy_ids=}")
                pairwise_policy_ids.append(policy_ids)

        if verbose:
            print(f"  - {len(pairwise_policy_ids)} comparisons to run")

        for policy_ids in pairwise_policy_ids:
            all_pairwise_comparisons.append(
                PWCParams(
                    env_id=env_id,
                    env_args_id=args_id,
                    env_args=env_args,
                    policy_ids=policy_ids,
                    output_file=output_file,
                    num_episodes=num_episodes,
                    seed=seed,
                    verbose=verbose,
                    run_num=run_num,
                )
            )
            run_num += 1

    if verbose:
        print(f"Running {len(all_pairwise_comparisons)} comparisons for {env_id=}.")

    return all_pairwise_comparisons


def run_pairwise_comparisons(
    env_id: str | None,
    env_args_id: str | None = None,
    num_episodes: int = 1000,
    output_dir: str | None = None,
    seed: int | None = None,
    n_procs: int | None = None,
    verbose: bool = False,
) -> str:
    """Run pairwise comparisons for all policies in an environment.

    Returns the directory where results are saved.
    """
    if env_id is None:
        env_ids = list(pga.get_all_envs())
        env_args_id = None
    else:
        env_ids = [env_id]
        env_args_id = env_args_id

    if output_dir is None:
        output_dir = os.path.join(
            BASE_RESULTS_DIR,
            "pairwise_comparison" + datetime.now().strftime("%d-%m-%y_%H-%M-%S"),
        )
    os.makedirs(output_dir, exist_ok=True)

    all_pairwise_comparisons = []
    for env_id in env_ids:
        all_pairwise_comparisons.extend(
            get_pairwise_comparison_params(
                env_id=env_id,
                env_args_id=env_args_id,
                output_dir=output_dir,
                num_episodes=num_episodes,
                seed=seed,
                verbose=verbose,
            )
        )

    print(f"Running {len(all_pairwise_comparisons)} comparisons using {n_procs=}.")
    print(f"Writing results to {output_dir=}")

    pw_args = [
        (params, len(all_pairwise_comparisons)) for params in all_pairwise_comparisons
    ]

    if n_procs is None or n_procs > 1:
        # limit number of cpus used by torch and numpy to 1 per process
        # note: setting this for torch should be enough but we set it for numpy too
        # since it changes the variable for OPENMP, MKL, etc. which numpy uses
        torch.set_num_threads(1)

    write_lock = mp.Lock()
    with mp.Pool(
        n_procs, initializer=_init_pool_processes, initargs=(write_lock,)
    ) as pool:
        pool.map(run_episodes, pw_args)

    print("All pairwise comparisons complete.")
    return output_dir


def load_pairwise_comparison_results(
    env_id: str,
    output_dir: str,
    env_args_id: str | None = None,
) -> pd.DataFrame:
    """Load pairwise comparison results for an environment."""
    env_output_dir = os.path.join(output_dir, env_id)
    output_file = os.path.join(env_output_dir, f"{env_args_id}.csv")
    if not os.path.exists(output_file):
        raise ValueError(
            f"Results do not exist for {env_id=}, {env_args_id=} in {output_dir=}."
        )
    return pd.read_csv(output_file)


def get_pairwise_returns_matrix(
    df: pd.DataFrame,
) -> Dict[str, Tuple[np.ndarray, List[str], List[str]]]:
    """Get pairwise returns matrix for each agent in environment."""
    num_episodes = df["num_episodes"].unique().tolist()[0]
    agent_ids = df["agent_id"].unique().tolist()
    agent_ids.sort()
    policies_per_agent = {
        i: df[df["agent_id"] == i]["policy_name"].unique().tolist() for i in agent_ids
    }
    co_teams_per_agent = {
        i: df[df["agent_id"] == i]["co_team_id"].unique().tolist() for i in agent_ids
    }
    co_teams_per_agent = {i: [] for i in agent_ids}
    for i in agent_ids:
        co_teams_i = df[df["agent_id"] == i]["co_team_id"].unique().tolist()
        # exclude non-homogenous teams
        for co_team in co_teams_i:
            pi_ids = co_team.replace("(", "").replace(")", "").split(",")
            if all(pi_j == pi_ids[0] for pi_j in pi_ids):
                co_teams_per_agent[i].append(co_team)

    for i in agent_ids:
        policies_per_agent[i].sort()
        co_teams_per_agent[i].sort()

    pw_returns_per_agent = {
        i: (
            np.zeros((3, len(policies_per_agent[i]), len(co_teams_per_agent[i]))),
            policies_per_agent[i],
            # store co-team labels for plotting, keep only first policy name for each
            # co-team
            [
                team_id.replace("(", "").replace(")", "").split(",")[0]
                for team_id in co_teams_per_agent[i]
            ],
        )
        for i in agent_ids
    }
    for i in agent_ids:
        pw_returns_i = pw_returns_per_agent[i][0]
        for (policy_idx, policy_name), (co_team_idx, co_team_id) in product(
            enumerate(policies_per_agent[i]), enumerate(co_teams_per_agent[i])
        ):
            if (
                len(
                    df[
                        (df["agent_id"] == i)
                        & (df["policy_name"] == policy_name)
                        & (df["co_team_id"] == co_team_id)
                    ]
                )
                == 0
            ):
                # no results for this policy/co-team matchup
                continue

            pw_returns_i[0, policy_idx, co_team_idx] = df[
                (df["agent_id"] == i)
                & (df["policy_name"] == policy_name)
                & (df["co_team_id"] == co_team_id)
            ]["episode_reward_mean"].values[0]
            pw_returns_i[1, policy_idx, co_team_idx] = df[
                (df["agent_id"] == i)
                & (df["policy_name"] == policy_name)
                & (df["co_team_id"] == co_team_id)
            ]["episode_reward_std"].values[0]
            # compute 95% CI for mean: 1.96*std/sqrt(N)
            pw_returns_i[2, policy_idx, co_team_idx] = (
                1.96 * pw_returns_i[1, policy_idx, co_team_idx] / np.sqrt(num_episodes)
            )
    return pw_returns_per_agent


def generate_pairwise_returns_plot(
    env_id: str,
    output_dir: str,
    env_args_id: str | None,
    pw_returns_per_agent: Dict[str, Tuple[np.ndarray, List[str], List[str]]],
    show: bool = True,
    save: bool = True,
    mean_only: bool = False,
) -> None:
    """Generate a plot of pairwise returns for an environment."""
    num_agents = len(pw_returns_per_agent)

    stats = ["mean"]
    if not mean_only:
        stats.extend(["std", "CI95"])

    for stat in stats:
        fig, axs = plt.subplots(
            nrows=1,
            ncols=num_agents,
            figsize=(6 * num_agents, 6),
            layout="constrained",
            squeeze=False,
        )
        for idx, i in enumerate(pw_returns_per_agent):
            pw_returns, policy_ids, co_teams_ids = pw_returns_per_agent[i]
            if stat == "mean":
                values = pw_returns[0]
            elif stat == "std":
                values = pw_returns[1]
            else:
                values = pw_returns[2]

            # only show first policy name for each co-team
            co_team_labels = [
                team_id.replace("(", "").replace(")", "").split(",")[0]
                for team_id in co_teams_ids
            ]

            sns.heatmap(
                values,
                ax=axs[0][idx],
                annot=True,
                fmt=".2f",
                cmap="YlGnBu",
                xticklabels=co_team_labels,
                yticklabels=policy_ids,
                square=True,
                annot_kws={"fontsize": min(8, 24 / np.sqrt(max(values.shape)))},
            )
            axs[0][idx].set_title(f"Return {stat} - agent `{i}`")

        if save:
            if env_id in output_dir:
                output_file = os.path.join(
                    output_dir, f"{env_args_id}_pairwise_returns_{stat}.svg"
                )
            else:
                output_file = os.path.join(
                    output_dir, f"{env_id}_{env_args_id}_pairwise_returns_{stat}.svg"
                )
            fig.savefig(output_file, bbox_inches="tight")

    if show:
        plt.show()


def plot_pairwise_comparison_results(
    output_dir: str,
    env_id: str | None,
    env_args_id: str | None = None,
    show: bool = True,
    save: bool = True,
    mean_only: bool = False,
):
    """Run return diversity analysis from pairwise comparison results.

    Results of analysis will be output to the results directory for each env.
    """
    all_envs = pga.get_all_envs()
    if env_id is None:
        env_ids = list(all_envs)
        env_args_id = None
    else:
        env_ids = [env_id]
        env_args_id = env_args_id

    for env_id in os.listdir(output_dir):
        if env_id not in env_ids:
            continue
        print(f"Plotting pairwise comparison results for {env_id=}")
        env_results_dir = os.path.join(output_dir, env_id)
        print(f"  results saved to {env_results_dir=}")

        env_result_files = [
            fname for fname in os.listdir(env_results_dir) if fname.endswith(".csv")
        ]

        env_args_map = pga.get_all_envs()[env_id]
        for args_id in env_args_map:
            if (env_args_id is not None and args_id != env_args_id) or (
                env_args_id is None and str(args_id) + ".csv" not in env_result_files
            ):
                continue

            print(f"  {args_id=}")
            df = load_pairwise_comparison_results(env_id, output_dir, args_id)

            if df["symmetric"].unique().tolist()[0]:
                # only keep results for one agent
                agent_ids = df["agent_id"].unique().tolist()
                df = df[df["agent_id"] == agent_ids[0]]

            pw_returns_per_agent = get_pairwise_returns_matrix(df)

            generate_pairwise_returns_plot(
                env_id=env_id,
                output_dir=env_results_dir,
                env_args_id=args_id,
                pw_returns_per_agent=pw_returns_per_agent,
                show=False,
                save=save,
                mean_only=mean_only,
            )

    if show:
        plt.show()



"""Synchronized vectorized environment class.

Based on Gymnasium Vectorized Environments:
https://github.com/Farama-Foundation/Gymnasium/blob/main/gymnasium/vector/vector_env.py
https://github.com/Farama-Foundation/Gymnasium/blob/main/gymnasium/vector/sync_vector_env.py

"""
from __future__ import annotations

from copy import deepcopy
from typing import Any, Callable, Dict, Iterable, List, Tuple

import numpy as np
from gymnasium.vector.utils import concatenate, create_empty_array
from gymnasium.vector.utils.spaces import batch_space

import posggym


class SyncVectorEnv(posggym.Env):
    """Vectorized environment that serially runs multiple environments.

    This implementation is based on the Gymnasium Vectorized Environments, modified to
    support multiple agents and the POSGGym API.

    A vectorized environment runs multiple independent copies of the same environment
    allowing for batched interactions. To prevent terminated environments waiting until
    all sub-environments are done (terminated or truncated), the vector environment
    autoresets sub-environments when they are done. As a result, the final observaiton
    and info are overwritten by the reset's observation and info. The final observation
    and info of the previous environment is stored in the info dictionary of each agent
    under the keys ``final_observation`` and ``final_info``. See :meth:`step` for more
    information.

    Vector Environments have additional attributes on top of `posggym.Env` as well as,
    slightly modified :attr:`observation_spaces` and :attr:`action_spaces` attributes:

    - :attr:`num_envs` - The number of sub-environments in the vector environment.
    - :attr:`observation_spaces` - The batched observation space of each agent of the
        vector environment.
    - :attr:`single_observation_spaces` - The observation space of each agent in a
        single sub-environment.
    - :attr:`action_spaces` - The batched action space of each agent of the  vector
        environment.
    - :attr:`single_action_spaces` - The action space of each agent in a  single
        sub-environment.


    Notes
    -----
    https://github.com/Farama-Foundation/Gymnasium/blob/main/gymnasium/vector/vector_env.py
    https://github.com/Farama-Foundation/Gymnasium/blob/main/gymnasium/vector/sync_vector_env.py

    """

    def __init__(
        self,
        env_fns: Iterable[Callable[[], posggym.Env]],
        copy: bool = True,
    ):
        """Initialize the vectorized environment.

        Arguments
        ---------
        env_fns:
            iterable of callable functions that create the environments.
        copy:
            If ``True``, then the :meth:`reset` and :meth:`step` methods return a
            copy of the observations.

        """
        self.env_fns = env_fns
        self.envs = [fn() for fn in env_fns]
        self.copy = copy
        self.is_vector_env = True

        self.metadata = self.envs[0].metadata
        self.model = self.envs[0].model

        self.num_envs = len(self.envs)

        self.single_observation_spaces = self.envs[0].observation_spaces
        self.single_action_spaces = self.envs[0].action_spaces

        self._observation_spaces = {
            i: batch_space(self.single_observation_spaces[i], n=self.num_envs)
            for i in self.single_observation_spaces
        }
        self._action_spaces = {
            i: batch_space(self.single_action_spaces[i], n=self.num_envs)
            for i in self.single_action_spaces
        }

        self._check_spaces()

        self.observations = {
            i: create_empty_array(
                self.single_observation_spaces[i], n=self.num_envs, fn=np.zeros
            )
            for i in self.single_observation_spaces
        }
        self._rewards = {
            i: np.zeros((self.num_envs,), dtype=np.float32)
            for i in self.single_observation_spaces
        }
        self._terminateds = {
            i: np.zeros((self.num_envs,), dtype=np.bool_)
            for i in self.single_observation_spaces
        }
        self._truncateds = {
            i: np.zeros((self.num_envs,), dtype=np.bool_)
            for i in self.single_observation_spaces
        }
        self._all_dones = np.zeros((self.num_envs,), dtype=np.bool_)
        self._actions = {i: None for i in self.single_action_spaces}

    def reset(
        self,
        *,
        seed: int | None | List[int] = None,
        options: Dict[str, Any] | None = None,
    ):
        """Reset all environments and return batch of initial observations and info."""
        if seed is None:
            seed = [None for _ in range(self.num_envs)]
        elif isinstance(seed, int):
            seed = [seed + i for i in range(self.num_envs)]
        assert len(seed) == self.num_envs

        self._all_dones[:] = False
        for i in self.single_observation_spaces:
            self._terminateds[i][:] = False
            self._truncateds[i][:] = False

        observations = {i: [] for i in self.single_observation_spaces}
        infos = {i: {} for i in self.single_observation_spaces}

        for env_num, (env, s) in enumerate(zip(self.envs, seed)):
            obs, info = env.reset(seed=s, options=options)
            for i in self.single_observation_spaces:
                observations[i].append(obs[i])
                infos[i] = self._add_info(infos[i], info[i], env_num)

        for i in self.single_observation_spaces:
            self.observations[i] = concatenate(
                self.single_observation_spaces[i],
                observations[i],
                self.observations[i],
            )

        return (deepcopy(self.observations) if self.copy else self.observations), infos

    def step(self, actions):
        """Take a step in all environments with the given actions.

        If any environment is done, then it is reset before taking the next step. The
        final observation and info of the previous environment is stored in the info
        dictionary under the keys ``final_observation`` and ``final_info``.


        Arguments
        ---------
        actions:
            dict mapping agent ID to batch of actions for that agent, with one action
            for each environment. So should be a dict of arrays or lists, with each
            array/list having length equal to the number of environments.

        Returns
        -------
        observations:
            dict mapping agent ID to batch of observations for that agent, with one
            observation for each environment. Each array having shape
            ``(num_envs,) + observation_space.shape``.
        rewards:
            dict mapping agent ID to batch of rewards for that agent, with one reward
            for each environment. Each entry is an array with  shape ``(num_envs,)``.
        terminateds:
            dict mapping agent ID to batch of termination signals for that agent, with
            one termination signal for each environment. Each entry is an array with
            shape ``(num_envs,)``.
        truncateds:
            dict mapping agent ID to batch of truncation signals for that agent, with
            one truncation signal for each environment. Each entry is an array with
            shape ``(num_envs,)``.
        all_dones:
            array of booleans, with one entry for each environment, indicating whether
            the environment is done. Shape is ``(num_envs,)``.
        infos:
            dict mapping agent ID to batch of info objects for that agent, with one
            info object for each environment. Each entry is a dict with one entry for


        The batched environment step results.

        """
        observations = {i: [] for i in self.single_observation_spaces}
        infos = {i: {} for i in self.single_observation_spaces}

        for env_num, env in enumerate(self.envs):
            action = {i: actions[i][env_num] for i in self.single_action_spaces}

            observation, rewards, terminateds, truncateds, all_done, info = env.step(
                action
            )

            if all_done:
                old_observation, old_info = observation, info
                observation, info = env.reset()
                for i in old_observation:
                    info[i]["final_observation"] = old_observation[i]
                    info[i]["final_info"] = old_info[i]

            for i in self.single_observation_spaces:
                observations[i].append(observation[i])
                infos[i] = self._add_info(infos[i], info[i], env_num)
                self._rewards[i][env_num] = rewards[i]
                self._terminateds[i][env_num] = terminateds[i]
                self._truncateds[i][env_num] = truncateds[i]

            self._all_dones[env_num] = all_done

        for i in self.single_observation_spaces:
            self.observations[i] = concatenate(
                self.single_observation_spaces[i],
                observations[i],
                self.observations[i],
            )

        return (
            (deepcopy(self.observations) if self.copy else self.observations),
            deepcopy(self._rewards) if self.copy else self._rewards,
            deepcopy(self._terminateds) if self.copy else self._terminateds,
            deepcopy(self._truncateds) if self.copy else self._truncateds,
            deepcopy(self._all_dones) if self.copy else self._all_dones,
            infos,
        )

    def render(self):
        return self.call("render")

    def close(self):
        """Close all environments."""
        for env in self.envs:
            env.close()

    def call(self, name: str, *args, **kwargs) -> Tuple:
        """Call a method on all environments and return the results."""
        results = []
        for env in self.envs:
            function = getattr(env, name)
            if callable(function):
                results.append(function(*args, **kwargs))
            else:
                results.append(function)
        return tuple(results)

    @property
    def possible_agents(self) -> Tuple[str, ...]:
        return self.envs[0].possible_agents

    @property
    def agents(self):
        return self.call("agents")

    @property
    def state(self):
        return self.call("state")

    @property
    def observation_spaces(self):
        return self._observation_spaces

    @property
    def action_spaces(self):
        return self._action_spaces

    def _add_info(self, infos: dict, info: dict, env_num: int) -> dict:
        """Add env info to the info dictionary of the vectorized environment.

        Given the `info` of a single environment add it to the `infos` dictionary
        which represents all the infos of the vectorized environment.
        Every `key` of `info` is paired with a boolean mask `_key` representing
        whether or not the i-indexed environment has this `info`.

        Arguments:
        ----------
        infos: dict
            the infos of the vectorized environment
        info: dict
            the info coming from the single environment
        env_num: int
            the index of the single environment

        Returns
        -------
        infos: dict
            the (updated) infos of the vectorized environment

        """
        for k in info:
            if k not in infos:
                info_array, array_mask = self._init_info_arrays(type(info[k]))
            else:
                info_array, array_mask = infos[k], infos[f"_{k}"]

            info_array[env_num], array_mask[env_num] = info[k], True
            infos[k], infos[f"_{k}"] = info_array, array_mask
        return infos

    def _init_info_arrays(self, dtype: type) -> Tuple[np.ndarray, np.ndarray]:
        """Initialize the info array.

        Initialize the info array. If the dtype is numeric the info array will have the
        same dtype, otherwise will be an array of `None`. Also, a boolean array of the
        same length is returned. It will be used for assessing which environment has
        info data.

        Arguments
        ---------
        dtype: type
            data type of the info coming from the env.

        Returns
        -------
        array: np.ndarray
            the initialized info array.
        array_mask: np.ndarray
            the initialized boolean array.

        """
        if dtype in [int, float, bool] or issubclass(dtype, np.number):
            array = np.zeros(self.num_envs, dtype=dtype)
        else:
            array = np.zeros(self.num_envs, dtype=object)
            array[:] = None
        array_mask = np.zeros(self.num_envs, dtype=bool)
        return array, array_mask

    def _check_spaces(self) -> bool:
        for env in self.envs:
            for i in self.single_observation_spaces:
                if (
                    i not in env.observation_spaces
                    or env.observation_spaces[i] != self.single_observation_spaces[i]
                ):
                    raise RuntimeError(
                        "Some environments have an observation space different from "
                        f"`{self.single_observation_spaces[i]}`. In order to batch "
                        "observations, the observation spaces from all environments "
                        "must be equal."
                    )

                if (
                    i not in env.action_spaces
                    or env.action_spaces[i] != self.single_action_spaces[i]
                ):
                    raise RuntimeError(
                        "Some environments have an action space different from "
                        f"`{self.single_action_spaces[i]}`. In order to batch actions, "
                        "the action spaces from all environments must be equal."
                    )

        return True



"""Functions and classes for rendering grid world environments."""
import abc
from typing import Dict, List, Optional, Tuple, Union

import numpy as np

from posggym.envs.grid_world.core import Coord, Direction, Grid
from posggym.error import DependencyNotInstalled

ColorTuple = Union[Tuple[int, int, int], Tuple[int, int, int, int]]

try:
    import pygame
except ImportError as e:
    raise DependencyNotInstalled(
        "pygame is not installed, run `pip install posggym[grid-world]`"
    ) from e


# (Main, Alternative) agent colors
AGENT_COLORS = [
    (pygame.colordict.THECOLORS["red"], pygame.colordict.THECOLORS["red3"]),
    (pygame.colordict.THECOLORS["blue"], pygame.colordict.THECOLORS["blue3"]),
    (pygame.colordict.THECOLORS["green"], pygame.colordict.THECOLORS["green3"]),
    (pygame.colordict.THECOLORS["purple"], pygame.colordict.THECOLORS["purple3"]),
    (pygame.colordict.THECOLORS["yellow"], pygame.colordict.THECOLORS["yellow3"]),
    (pygame.colordict.THECOLORS["cyan"], pygame.colordict.THECOLORS["cyan3"]),
    (pygame.colordict.THECOLORS["hotpink"], pygame.colordict.THECOLORS["hotpink3"]),
    (pygame.colordict.THECOLORS["orange"], pygame.colordict.THECOLORS["orange3"]),
]


def get_agent_color(agent_id: str) -> Tuple[ColorTuple, ColorTuple]:
    """Get color for agent."""
    return AGENT_COLORS[int(agent_id) % len(AGENT_COLORS)]


def get_color(color_name: str) -> ColorTuple:
    """Get color from name."""
    return pygame.colordict.THECOLORS[color_name]


def load_img_file(img_path: str, cell_size: Tuple[int, int]):
    """Load an image from file and scale it to cell size."""
    return pygame.transform.scale(pygame.image.load(img_path), cell_size)


def get_default_font_size(cell_size: Tuple[int, int]) -> int:
    """Get the default font size based on cell size."""
    return cell_size[1] // 2


def load_font(font_name: str, font_size: int) -> pygame.font.Font:
    """Load font."""
    return pygame.font.SysFont(font_name, font_size)


class GWObject(abc.ABC):
    """An object in the grid world."""

    def __init__(
        self,
        coord: Coord,
        cell_size: Tuple[int, int],
    ):
        self.coord = coord
        self.cell_size = cell_size

    @property
    def pos(self) -> Tuple[int, int]:
        """The (x, y) position of the object on render surface."""
        return (self.coord[0] * self.cell_size[0], self.coord[1] * self.cell_size[1])

    @abc.abstractmethod
    def render(self, surface: pygame.Surface):
        """Render object on surface."""


class GWRectangle(GWObject):
    """A rectangle in the grid world."""

    def __init__(
        self,
        coord: Coord,
        cell_size: Tuple[int, int],
        color: ColorTuple,
    ):
        super().__init__(coord, cell_size)
        self.color = color

    def render(self, surface: pygame.Surface):
        rect = (*self.pos, *self.cell_size)
        surface.fill(self.color, rect=rect)


class GWTriangle(GWObject):
    """A equilateral triangle in the grid world."""

    def __init__(
        self,
        coord: Coord,
        cell_size: Tuple[int, int],
        color: ColorTuple,
        facing_dir: Direction,
    ):
        super().__init__(coord, cell_size)
        self.color = color
        self.facing_dir = facing_dir

    def _get_triangle_by_direction(self, facing_dir: Direction):
        # amount to shrink points away from cell edges, so they fit better
        delta = (int(0.05 * self.cell_size[0]), int(0.05 * self.cell_size[1]))

        max_pos = (
            self.pos[0] + self.cell_size[0] - delta[0],
            self.pos[1] + self.cell_size[1] - delta[1],
        )
        min_pos = (self.pos[0] + delta[0], self.pos[1] + delta[1])

        if facing_dir == Direction.NORTH:
            points = (
                (min_pos[0] + self.cell_size[0] // 2, min_pos[1]),
                (min_pos[0], max_pos[1]),
                (max_pos[0], max_pos[1]),
            )
        elif facing_dir == Direction.SOUTH:
            points = (
                (min_pos[0] + self.cell_size[0] // 2, max_pos[1]),
                (min_pos[0], min_pos[1]),
                (max_pos[0], min_pos[1]),
            )
        elif facing_dir == Direction.EAST:
            points = (
                (max_pos[0], min_pos[1] + self.cell_size[1] // 2),
                (min_pos[0], min_pos[1]),
                (min_pos[0], max_pos[1]),
            )
        elif facing_dir == Direction.WEST:
            points = (
                (min_pos[0], min_pos[1] + self.cell_size[1] // 2),
                (max_pos[0], min_pos[1]),
                (max_pos[0], max_pos[1]),
            )

        return points

    def render(self, surface: pygame.Surface):
        points = self._get_triangle_by_direction(self.facing_dir)
        pygame.draw.polygon(surface, self.color, points)


class GWCircle(GWObject):
    """A circle in the grid world."""

    def __init__(
        self,
        coord: Coord,
        cell_size: Tuple[int, int],
        color: ColorTuple,
    ):
        super().__init__(coord, cell_size)
        self.color = color

    def render(self, surface: pygame.Surface):
        radius = int(0.95 * self.cell_size[0]) // 2
        center = (
            self.pos[0] + self.cell_size[0] // 2,
            self.pos[1] + self.cell_size[1] // 2,
        )
        pygame.draw.circle(surface, self.color, center, radius)


class GWHighlight(GWObject):
    """A transparent rectangle for highlighting a cell in the grid world."""

    def __init__(self, coord: Coord, cell_size: Tuple[int, int], alpha: float = 0.25):
        super().__init__(coord, cell_size)
        self.alpha = alpha
        self.surface = pygame.Surface(cell_size, pygame.SRCALPHA)
        self.surface.fill((255, 255, 255, int(alpha * 255)))

    def render(self, surface: pygame.Surface):
        surface.blit(self.surface, self.pos)


class GWImage(GWObject):
    """An image object in the grid world."""

    def __init__(
        self,
        coord: Coord,
        cell_size: Tuple[int, int],
        img: pygame.Surface,
    ):
        super().__init__(coord, cell_size)
        self.img = img

    def render(self, surface: pygame.Surface):
        surface.blit(self.img, self.pos)


class GWText(GWObject):
    """An text object in the grid world."""

    def __init__(
        self,
        coord: Coord,
        cell_size: Tuple[int, int],
        text: str,
        font: pygame.font.Font,
    ):
        super().__init__(coord, cell_size)
        self.text = text
        self.font = font

    def render(self, surface: pygame.Surface):
        text_img = self.font.render(self.text, True, (255, 255, 255), (0, 0, 0))
        # +1 so text is in offset from corner of cell
        text_pos = (self.pos[0] + 1, self.pos[1] + 1)
        surface.blit(text_img, text_pos)


class GWImageAndText(GWObject):
    """A combined image and text grid-world object."""

    def __init__(
        self,
        coord: Coord,
        cell_size: Tuple[int, int],
        img: pygame.Surface,
        text: str,
        font: pygame.font.Font,
    ):
        super().__init__(coord, cell_size)
        self.img = img
        self.text = text
        self.font = font

    def render(self, surface: pygame.Surface):
        surface.blit(self.img, self.pos)

        text_img = self.font.render(self.text, True, (255, 255, 255), (0, 0, 0))
        # +1 so text is in offset from corner of cell
        text_pos = (self.pos[0] + 1, self.pos[1] + 1)
        surface.blit(text_img, text_pos)


class GWRenderer:
    """Handles generating grid world renders."""

    def __init__(
        self,
        render_mode: str,
        grid: Grid,
        render_fps: int = 30,
        env_name: str = "",
        bg_color: ColorTuple = (0, 0, 0),
        grid_line_color: ColorTuple = (255, 255, 255),
        block_color: ColorTuple = (131, 139, 139),
    ):
        self.render_mode = render_mode
        self.grid = grid
        self.render_fps = render_fps

        self.bg_color = bg_color
        self.grid_line_color = grid_line_color
        self.block_color = block_color

        self.window_size = (min(64 * grid.width, 512), min(64 * grid.height, 512))
        self.cell_size = (
            self.window_size[0] // grid.width,
            self.window_size[1] // grid.height,
        )

        self.blocks = [
            GWRectangle(coord, self.cell_size, self.block_color)
            for coord in grid.block_coords
        ]
        # list of static objects user can add to
        self.static_objects: List[GWObject] = []

        pygame.init()
        if render_mode == "human":
            pygame.display.init()
            pygame.display.set_caption(env_name)
            self.window_surface = pygame.display.set_mode(
                self.window_size, pygame.SRCALPHA
            )
        else:
            assert render_mode.startswith("rgb_array")
            pygame.font.init()
            self.window_surface = pygame.Surface(self.window_size, pygame.SRCALPHA)

        self.clock = pygame.time.Clock()

    def _reset_surface(self):
        self.window_surface.fill(self.bg_color)
        for y in range(self.grid.height):
            for x in range(self.grid.width):
                pos = (x * self.cell_size[0], y * self.cell_size[1])
                rect = (*pos, *self.cell_size)
                pygame.draw.rect(self.window_surface, self.grid_line_color, rect, 1)

        for block in self.blocks:
            block.render(self.window_surface)

        for static_object in self.static_objects:
            static_object.render(self.window_surface)

    def reset_blocks(self):
        """Reset the blocks."""
        self.blocks = [
            GWRectangle(coord, self.cell_size, self.block_color)
            for coord in self.grid.block_coords
        ]

    def render(
        self, objects: List[GWObject], observed_coords: Optional[List[Coord]] = None
    ) -> Optional[np.ndarray]:
        """Generate Grid-World render."""
        self._reset_surface()
        for obj in objects:
            obj.render(self.window_surface)

        if observed_coords is None:
            observed_coords = []

        highlight_obj = GWHighlight((0, 0), self.cell_size)
        for coord in observed_coords:
            highlight_obj.coord = coord
            highlight_obj.render(self.window_surface)

        if self.render_mode == "human":
            pygame.event.pump()
            pygame.display.update()
            self.clock.tick(self.render_fps)
            return None

        return np.transpose(
            np.array(pygame.surfarray.pixels3d(self.window_surface)), axes=(1, 0, 2)
        )

    def render_agents(
        self,
        objects: List[GWObject],
        agent_coords_and_dirs: Dict[str, Tuple[Coord, Direction]],
        agent_obs_dims: Union[int, Tuple[int, int, int, int]],
        observed_coords: Optional[List[Coord]] = None,
        agent_obs_mask: Optional[List[Coord]] = None,
    ) -> Dict[str, np.ndarray]:
        """Generate environment and agent-centric grid-world renders."""
        if agent_obs_mask is None:
            agent_obs_mask = []
        if observed_coords is None:
            observed_coords = []
        if isinstance(agent_obs_dims, int):
            d = agent_obs_dims
            agent_obs_dims = (d, d, d, d)

        self._reset_surface()
        for obj in objects:
            obj.render(self.window_surface)

        env_array = np.array(pygame.surfarray.pixels3d(self.window_surface))

        array_dict: Dict[str, np.ndarray] = {}
        for i, (coord, facing_dir) in agent_coords_and_dirs.items():
            # 1. get agent's view of env
            # (min_col, max_col, min_row, max_row) of coords in grid that agent observed
            view_coords_rect = self.grid.get_rectangular_bounds(
                coord, facing_dir, agent_obs_dims
            )
            view_pos_rect = (
                view_coords_rect[0] * self.cell_size[0],
                (view_coords_rect[1] + 1) * self.cell_size[0],
                view_coords_rect[2] * self.cell_size[1],
                (view_coords_rect[3] + 1) * self.cell_size[1],
            )
            agent_array = env_array[
                view_pos_rect[0] : view_pos_rect[1], view_pos_rect[2] : view_pos_rect[3]
            ]

            # 2. pad view with blocks so agent's view matches their obs_dims
            pad_coords = self.grid.get_rectangular_padding(
                coord, facing_dir, agent_obs_dims
            )
            pad_pos = [
                (
                    pad_coords[0][0] * self.cell_size[0],
                    pad_coords[0][1] * self.cell_size[0],
                ),
                (
                    pad_coords[1][0] * self.cell_size[1],
                    pad_coords[1][1] * self.cell_size[1],
                ),
                # padding for color axes (i.e no padding)
                (0, 0),
            ]
            agent_array = np.pad(
                agent_array,
                pad_width=pad_pos,
                mode="constant",
                constant_values=[
                    (self.block_color[0], self.block_color[0]),
                    (self.block_color[1], self.block_color[1]),
                    (self.block_color[2], self.block_color[2]),
                ],
            )

            # 3. apply masking if necessary
            for coord in agent_obs_mask:
                pos = (coord[0] * self.cell_size[0], coord[1] * self.cell_size[1])
                agent_array[
                    pos[0] : pos[0] + self.cell_size[0],
                    pos[1] : pos[1] + self.cell_size[1],
                ] = self.bg_color[:3]

            # 3. rotate agent's array so it is ego-centric
            agent_array = np.transpose(agent_array, axes=(1, 0, 2))
            # just so happens int value of direction lines up with # rotations needed
            agent_array = np.rot90(agent_array, k=int(facing_dir))

            array_dict[i] = agent_array

        # Add obs highlighting after generating agent-centric views
        highlight_obj = GWHighlight((0, 0), self.cell_size)
        for coord in observed_coords:
            highlight_obj.coord = coord
            highlight_obj.render(self.window_surface)

        # finally add env image, including highlighting
        array_dict["env"] = np.transpose(
            np.array(pygame.surfarray.pixels3d(self.window_surface)), axes=(1, 0, 2)
        )

        return array_dict

    def close(self):
        """Close renderer and perform any necessary cleanup."""
        pygame.display.quit()
        pygame.quit()



"""Wrapper for incorporating posggym.agents as part of the environment."""
from typing import Callable, Dict, List, Tuple

from gymnasium import spaces

import posggym
import posggym.agents as pga


class AgentEnvWrapper(posggym.Wrapper):
    """This wrapper incorporates posggym.agent policies as part of the environment.

    This wrapper makes it so that specified agents within the environment have their
    actions determined internally by a posggym.agent policy. The environment wrapper
    will only return observations, rewards, etc, for agents which are not controlled.

    Arguments
    ---------
    env : posggym.Env
        The environment to apply the wrapper
    agent_fn : Callable[[posggym.POSGModel], Dict[str, pga.Policy]]
        A function which returns a policy for each agent ID in the environment to be
        controlled as part of the environment. Should always return a policy for the
        same set of agent IDs.
    """

    def __init__(
        self,
        env: posggym.Env,
        agent_fn: Callable[[posggym.POSGModel], Dict[str, pga.Policy]],
    ):
        """Initializes the wrapper."""
        super().__init__(env)
        self.agent_fn = agent_fn
        self.policies = agent_fn(self.model)
        self.controlled_agents = list(self.policies)
        self.last_obs = {}
        self.last_terminateds = {}

    @property
    def possible_agents(self) -> Tuple[str, ...]:
        return tuple(
            i for i in super().possible_agents if i not in self.controlled_agents
        )

    @property
    def agents(self) -> List[str]:
        return [i for i in super().agents if i not in self.controlled_agents]

    @property
    def action_spaces(self) -> Dict[str, spaces.Space]:
        return {
            i: act_space
            for i, act_space in super().action_spaces.items()
            if i not in self.controlled_agents
        }

    @property
    def observation_spaces(self) -> Dict[str, spaces.Space]:
        return {
            i: obs_space
            for i, obs_space in super().observation_spaces.items()
            if i not in self.controlled_agents
        }

    @property
    def reward_ranges(self) -> Dict[str, Tuple[float, float]]:
        return {
            i: rew_range
            for i, rew_range in super().reward_ranges.items()
            if i not in self.controlled_agents
        }

    def reset(self, **kwargs):
        obs, infos = super().reset(**kwargs)
        self.policies = self.agent_fn(self.model)

        assert set(self.policies) == set(
            self.controlled_agents
        ), "Agent IDs must be consistent across resets."

        for policy in self.policies.values():
            policy.reset()

        self.last_obs = {i: obs[i] for i in self.controlled_agents if i in obs}
        self.last_terminateds = {i: False for i in self.controlled_agents}
        return {i: o for i, o in obs.items() if i not in self.controlled_agents}, {
            i: info for i, info in infos.items() if i not in self.controlled_agents
        }

    def step(self, actions):
        for i, policy in self.policies.items():
            assert i not in actions, f"Agent {i} is controlled by the environment."
            if not self.last_terminateds[i]:
                actions[i] = policy.step(self.last_obs[i])

        obs, rewards, terminated, truncated, dones, infos = super().step(actions)
        self.last_obs = {i: obs[i] for i in self.controlled_agents if i in obs}
        self.last_terminateds = {
            i: terminated[i] for i in self.controlled_agents if i in terminated
        }
        return (
            {i: obs[i] for i in obs if i not in self.controlled_agents},
            {i: rewards[i] for i in rewards if i not in self.controlled_agents},
            {i: terminated[i] for i in terminated if i not in self.controlled_agents},
            {i: truncated[i] for i in truncated if i not in self.controlled_agents},
            dones,
            {i: infos[i] for i in infos if i not in self.controlled_agents},
        )



"""The Pursuit-Evasion World World Environment."""
import math
from itertools import product
from typing import (
    Any,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

import numpy as np
from gymnasium import spaces

import posggym.model as M
from posggym import logger
from posggym.core import DefaultEnv
from posggym.envs.continuous.core import (
    CollisionType,
    Coord,
    FloatCoord,
    PMBodyState,
    SquareContinuousWorld,
    clip_actions,
    generate_interior_walls,
)
from posggym.utils import seeding


class PEState(NamedTuple):
    """Environment state in Pursuit Evastion problem."""

    evader_state: np.ndarray
    pursuer_state: np.ndarray
    evader_start_coord: np.ndarray
    pursuer_start_coord: np.ndarray
    evader_goal_coord: np.ndarray
    min_goal_dist: int  # for evader


PEAction = np.ndarray
PEObs = np.ndarray


class PursuitEvasionContinuousEnv(DefaultEnv):
    """The Pursuit-Evasion Continuous World Environment.

    An adversarial continuous world problem involving two agents: an evader and a
    pursuer. The evader's goal is to reach a goal location, on the other side of
    the world, while the goal of the pursuer is to spot the evader before it reaches
    it's goal. The evader is considered caught if it is observed by the pursuer, or
    occupies the same location. The evader and pursuer have knowledge of each others
    starting locations, however only the evader has knowledge of it's goal location.
    The pursuer only knowns that the evader's goal location is somewhere on the opposite
    side of the world to the evaders start location.

    This environment requires each agent to reason about the which path the other agent
    will take through the environment.

    Possible Agents
    ---------------
    - Evader = `"0"`
    - Pursuer = `"1"`

    State Space
    -----------
    Each state is made up of:

    0. the state of the evader
    1. the state of the pursuer
    2. the `(x, y)` coordinate of the evader's start location
    3. the `(x, y)` coordinate of the pursuer's start location
    4. the `(x, y)` coordinate of the evader's goal
    5. the minimum distance to it's goal along the shortest discrete path achieved
       by the evader in the current episode (this is needed to correctly reward
       the agent for making progress.)

    Both the evader and pursuer state consist of their:

    - `(x, y)` coordinate
    - angle/yaw (in radians)
    - velocity in x and y directions
    - angular velocity (in radians)

    Action Space
    ------------
    Each agent's actions is made up of two parts. The first action component specifies
    the angular velocity in `[-pi/4, pi/4]`, and the second component specifies the
    linear velocity in `[0, 1]`.

    Observation Space
    -----------------
    Each agent observes the local world in a `fov` radian cone in front of
    themselves. This is achieved by a series of `n_sensors` sensor lines starting
    at the agent which extend up to `max_obs_distance` away from the agent.
    For each sensor line the agent observes the closest entity along the line,
    specifically if there is a wall or the other agent. Along with the sensor
    readings each agent also observes whether they hear the other agent within
    a circle with radius ``HEARING_DIST = 4.0` around themselves, the `(x, y)`
    coordinate of the evader's start location and the `(x, y)` coordinate of the
    pursuer's start location. The evader also observes  the `(x, y)` coordinate of
    their goal location, while the pursuer receives a value of `(0, 0` for
    this feature.

    This table enumerates the first part of the observation space:

    | Index: start          | Description                          |  Values   |
    | :-------------------: | :----------------------------------: | :-------: |
    | 0                     | Wall distance                        | [0, d]    |
    | n_sensors             | Other agent distance                 | [0, d]    |
    | 2 * n_sensors         | Other agent heard                    | [0, 1]    |
    | 2 * n_sensors + 1     | Evader start coordinates             | [0, s]    |
    | 2 * n_sensors + 3     | Pursuer start coordinates            | [0, s]    |
    | 2 * n_sensors + 5     | Evader goal coordinates              | [0, s]    |

    Where `d = max_obs_distance` and `s = world.size`

    If an entity is not observed by a sensor (i.e. it's not within `max_obs_distance`
    or is not the closest entity to the observing agent along the line), The distance
    reading will be `max_obs_distance`.

    Note, the goal and start coordinate observations do not change during a single
    episode, but they do change between episodes.

    Rewards
    -------
    The environment is zero-sum with the pursuer receiving the negative of the evader
    reward. Additionally, rewards are by default normalized so that returns are bounded
    between `-1` and `1` (this can be disabled by the `normalize_reward` parameter).

    The evader receives a reward of `1` for reaching it's goal location and a
    reward of `-1` if it gets captured. Additionally, the evader receives a small
    reward of `0.01` each time it's minimum distance achieved to it's goal along the
    shortest path decreases for the current episode. This is to make it so the
    environment is no longer sparesely rewarded and helps with exploration and learning
    (it can be disabled by the `use_progress_reward` parameter.)

    Dynamics
    --------
    Actions are deterministic and will change the agents direction and velocity in the
    direction they are facing. The velocity range of an agent is [0, 1]; agent's cannot
    move backwards (have negative velocity).

    Starting State
    --------------
    At the start of each episode the start location of the evader is selected at random
    from all possible start locations. The evader's goal location is then chosen
    randomly from the set of available goal locations given the evaders start location
    (in the default maps the goal location is always on the opposite side of the map
    from a start location). The pursuers start location is similarly chosen from it's
    set of possible start locations.

    Episode End
    -----------
    An episode ends when either the evader is seen or touched by the pursuer or the
    evader reaches it's goal. By default a `max_episode_steps` limit of `200` steps is
    also set. This may need to be adjusted when using larger worlds (this can be done
    by manually specifying a value for `max_episode_steps` when creating the environment
    with `posggym.make`).

    Arguments
    ---------

    - `world` - the world layout to use. This can either be a string specifying one of
         the supported worlds (see SUPPORTED_WORLDS), or a custom :class:`PEWorld`
         object (default = `"16x16"`).
    - `max_obs_distance` - the maximum distance from the agent that each agent's field
        of vision extends. If `None` then sets this to 1/3 of world size
        (default = `None`).
    - `fov` - the field of view of the agent in radians. This will determine the
        angle of the cone in front of the agent which it can see. The FOV will be
        relative to the angle of the agent (default = `pi / 3`).
    - `n_sensors` - the number of sensor lines eminating from the agent within their
        FOV. The agent will observe at `n_sensors` equidistance intervals over
        `[-fov / 2, fov / 2]` (default = `16`).
    - `normalize_reward` - whether to normalize both agents' rewards to be between `-1`
        and `1` (default = 'True`)
    - `use_progress_reward` - whether to reward the evader agent for making progress
        towards it's goal. If False the evader will only be rewarded when it reaches
        it's goal, making it a sparse reward problem (default = 'True`).

    Available variants
    ------------------

    The PursuitEvasionContinuous environment comes with a number of pre-built world
    layouts which can be passed as an argument to `posggym.make`, to create different
    worlds.

    | World name       | World size |
    |------------------|------------|
    | `8x8`            |  8x8       |
    | `16x16`          |  16x16     |
    | `32x32`          |  32x32     |

    For example to use the PursuitEvasionContinuous environment with the `32x32` world
    layout, and episode step limit of 200, and the default values for the other
    parameters you would use:

    ```python
    import posggym
    env = posggym.make(
        'PursuitEvasionContinuous-v0',
        max_episode_steps=200,
        world="32x32",
    )
    ```

    References
    ----------
    - [This Pursuit-Evasion implementation is directly inspired by the problem] Seaman,
      Iris Rubi, Jan-Willem van de Meent, and David Wingate. 2018. “Nested Reasoning
      About Autonomous Agents Using Probabilistic Programs.”
      ArXiv Preprint ArXiv:1812.01569.
    - Schwartz, Jonathon, Ruijia Zhou, and Hanna Kurniawati. "Online Planning for
      Interactive-POMDPs using Nested Monte Carlo Tree Search." In 2022 IEEE/RSJ
      International Conference on Intelligent Robots and Systems (IROS), pp. 8770-8777.
      IEEE, 2022.

    """

    metadata = {
        "render_modes": ["human", "rgb_array"],
        "render_fps": 15,
    }

    def __init__(
        self,
        world: Union[str, "PEWorld"] = "16x16",
        max_obs_distance: Optional[float] = None,
        fov: float = np.pi / 3,
        n_sensors: int = 16,
        normalize_reward: bool = True,
        use_progress_reward: bool = True,
        render_mode: Optional[str] = None,
        **kwargs,
    ):
        model = PursuitEvasionContinuousModel(
            world,
            max_obs_distance=max_obs_distance,
            normalize_reward=normalize_reward,
            use_progress_reward=use_progress_reward,
            fov=fov,
            n_sensors=n_sensors,
            **kwargs,
        )
        super().__init__(model, render_mode=render_mode)
        self.window_surface = None
        self.blocked_surface = None
        self.clock = None
        self.window_size = 600
        self.draw_options = None
        self.world = None
        self.fov = fov

    def reset(
        self, *, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, M.ObsType], Dict[str, Dict]]:
        # reset renderer since goal location can change between episodes
        self._renderer = None
        return super().reset(seed=seed, options=options)

    def render(self):
        if self.render_mode is None:
            assert self.spec is not None
            logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. posggym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return
        return self._render_img()

    def _render_img(self):
        import pygame
        from pymunk import Transform, pygame_util

        model = cast(PursuitEvasionContinuousModel, self.model)
        state = cast(PEState, self.state)
        scale_factor = self.window_size / model.world.size

        if self.window_surface is None:
            pygame.init()
            if self.render_mode == "human":
                pygame.display.init()
                pygame.display.set_caption(self.__class__.__name__)
                self.window_surface = pygame.display.set_mode(
                    (self.window_size, self.window_size)
                )
            else:
                self.window_surface = pygame.Surface(
                    (self.window_size, self.window_size)
                )
            # Turn off alpha since we don't use it.
            self.window_surface.set_alpha(None)

        if self.clock is None:
            self.clock = pygame.time.Clock()

        if self.draw_options is None:
            pygame_util.positive_y_is_up = False
            self.draw_options = pygame_util.DrawOptions(self.window_surface)
            self.draw_options.transform = Transform.scaling(scale_factor)
            # don't render collision lines
            self.draw_options.flags = (
                pygame_util.DrawOptions.DRAW_SHAPES
                | pygame_util.DrawOptions.DRAW_CONSTRAINTS
            )

        if self.world is None:
            # get copy of model world, so we can use it for rendering without
            # affecting the original
            self.world = model.world.copy()

        if self.blocked_surface is None:
            self.blocked_surface = pygame.Surface((self.window_size, self.window_size))
            self.blocked_surface.fill(pygame.Color("white"))
            cell_size = scale_factor
            for coord in self.world.blocked_coords:
                rect = (
                    round(coord[0] * cell_size),
                    round(coord[1] * cell_size),
                    round(cell_size),
                    round(cell_size),
                )
                self.blocked_surface.fill(self.world.BLOCK_COLOR, rect=rect)

        self.world.set_entity_state("pursuer", state.pursuer_state)
        self.world.set_entity_state("evader", state.evader_state)

        # Need to do this for space to update with changes
        self.world.space.step(0.0001)

        # reset screen
        self.window_surface.fill(pygame.Color("white"))

        # add blocks
        self.window_surface.blit(self.blocked_surface, (0, 0))

        # draw sensor lines
        n_sensors = model.n_sensors
        for idx, p_state in [
            (model.EVADER_IDX, state.evader_state),
            (model.PURSUER_IDX, state.pursuer_state),
        ]:
            obs_i = self._last_obs[str(idx)]
            x, y, agent_angle = p_state[:3]

            angles = np.linspace(
                -self.fov / 2, self.fov / 2, n_sensors, endpoint=False, dtype=np.float32
            )
            for k in range(n_sensors):
                dist = min([obs_i[k], obs_i[n_sensors + k]])
                angle = angles[k] + agent_angle
                end_x = x + dist * math.cos(angle)
                end_y = y + dist * math.sin(angle)
                scaled_start = (int(x * scale_factor), int(y * scale_factor))
                scaled_end = int(end_x * scale_factor), (end_y * scale_factor)
                pygame.draw.line(
                    self.window_surface,
                    pygame.Color("red"),
                    scaled_start,
                    scaled_end,
                )

        x, y = state.evader_goal_coord
        scaled_center = (int(x * scale_factor), int(y * scale_factor))
        scaled_r = int(model.world.agent_radius * scale_factor)
        pygame.draw.circle(
            self.window_surface,
            pygame.Color(model.GOAL_COLOR),
            scaled_center,
            scaled_r,
        )

        self.world.space.debug_draw(self.draw_options)

        if self.render_mode == "human":
            pygame.event.pump()
            pygame.display.update()
            self.clock.tick(self.metadata["render_fps"])
            return None

        return np.transpose(
            np.array(pygame.surfarray.pixels3d(self.window_surface)), axes=(1, 0, 2)
        )

    def close(self) -> None:
        if self.window_surface is not None:
            import pygame

            pygame.display.quit()
            pygame.quit()


class PursuitEvasionContinuousModel(M.POSGModel[PEState, PEObs, PEAction]):
    """Continuous Pursuit-Evasion Model.

    Arguments
    ---------
    world : str, PEWorld
        the world layout to use. This can either be a string specifying one of
        the supported worlds (see SUPPORTED_WORLDS), or a custom :class:`PEWorld`
        object.
    max_obs_distance : float, optional
        the maximum distance from the agent that each agent's field of vision extends.
        If `None` the distance will be set to 1/3 of the world's size.
    fov : float
        the field of view of the agent in radians. This will determine the angle of the
        cone in front of the agent which it can see. The FOV will be relative to the
        angle of the agent.
    n_sensors : int
        the number of sensor lines eminating from the agent within their FOV. The agent
        will observe at `n_sensors` equidistance intervals over `[-fov / 2, fov / 2]`.
    normalize_reward : bool
        whether to normalize both agents' rewards to be between `-1` and `1`
    use_progress_reward : bool
        whether to reward the evader agent for making progress towards it's goal. If
        False the evader will only be rewarded when it reaches it's goal, making it a
        sparse reward problem.

    """

    NUM_AGENTS = 2

    EVADER_IDX = 0
    PURSUER_IDX = 1

    # Evader-centric rewards
    R_PROGRESS = 0.01  # Reward making progress toward goal
    R_EVASION = 1.0  # Reward for reaching goal (capture reward is -R_EVASION)

    HEARING_DIST = 4

    EVADER_COLOR = (55, 155, 205, 255)  # Blueish
    PURSUER_COLOR = (255, 55, 55, 255)  # Purpleish
    GOAL_COLOR = (55, 255, 55, 255)  # Greenish

    def __init__(
        self,
        world: Union[str, "PEWorld"],
        max_obs_distance: Optional[float] = None,
        fov: float = np.pi / 3,
        n_sensors: int = 16,
        normalize_reward: bool = True,
        use_progress_reward: bool = True,
    ):
        assert 0 < fov < 2 * np.pi, "fov must be in (0, 2 * pi)"
        assert n_sensors > 0, "n_sensors must be positive"

        if isinstance(world, str):
            assert world in SUPPORTED_WORLDS, (
                f"Unsupported world name '{world}'. If world is a string it must be "
                f"one of: {SUPPORTED_WORLDS.keys()}."
            )
            world = SUPPORTED_WORLDS[world]()

        self.world = world

        if max_obs_distance is None:
            max_obs_distance = self.world.size / 3
        assert max_obs_distance > 0, "max_obs_distance must be positive"
        self.max_obs_distance = max_obs_distance
        self._normalize_reward = normalize_reward
        self._use_progress_reward = use_progress_reward
        self.fov = fov

        self._max_sp_distance = self.world.get_max_shortest_path_distance()
        self._max_raw_return = self.R_EVASION
        if self._use_progress_reward:
            self._max_raw_return += (self._max_sp_distance + 1) * self.R_PROGRESS
        self._min_raw_return = -self._max_raw_return

        def _coord_space():
            return spaces.Box(
                low=np.array([0, 0], dtype=np.float32),
                high=np.array([self.world.size, self.world.size], dtype=np.float32),
            )

        self.possible_agents = tuple(str(i) for i in range(self.NUM_AGENTS))
        # e_state, p_state, e_start_coord, p_start_coord, e_goal_coord, min_sp
        self.state_space = spaces.Tuple(
            (
                PMBodyState.get_space(self.world.size),
                PMBodyState.get_space(self.world.size),
                _coord_space(),
                _coord_space(),
                _coord_space(),
                spaces.Discrete(self._max_sp_distance + 1),
            )
        )

        # can turn by up to 45 degrees per timestep
        self.dyaw_limit = math.pi / 4
        self.action_spaces = {
            i: spaces.Box(
                low=np.array([-self.dyaw_limit, 0.0], dtype=np.float32),
                high=np.array([self.dyaw_limit, 1.0], dtype=np.float32),
            )
            for i in self.possible_agents
        }

        self.obs_dist = self.max_obs_distance
        self.n_sensors = n_sensors
        self.sensor_obs_dim = self.n_sensors * 2
        self.obs_dim = self.sensor_obs_dim + 7
        sensor_low = [0.0] * self.sensor_obs_dim
        sensor_high = [self.max_obs_distance] * self.sensor_obs_dim
        size = self.world.size
        self.observation_spaces = {
            i: spaces.Box(
                low=np.array([*sensor_low, 0, 0, 0, 0, 0, 0, 0], dtype=np.float32),
                high=np.array(
                    [*sensor_high, 1, size, size, size, size, size, size],
                    dtype=np.float32,
                ),
                dtype=np.float32,
            )
            for i in self.possible_agents
        }
        self.is_symmetric = False

        # Add physical entities to the world
        self.world.add_entity("pursuer", None, color=self.PURSUER_COLOR)
        self.world.add_entity("evader", None, color=self.EVADER_COLOR)

    @property
    def reward_ranges(self) -> Dict[str, Tuple[float, float]]:
        max_reward = self.R_EVASION
        if self._use_progress_reward:
            max_reward += self.R_PROGRESS
        if self._normalize_reward:
            max_reward = self._get_normalized_reward(max_reward)
        return {i: (-max_reward, max_reward) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: PEState) -> List[str]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> PEState:
        evader_coord = self.rng.choice(self.world.evader_start_coords)
        evader_state = np.zeros(PMBodyState.num_features(), dtype=np.float32)
        evader_state[:2] = evader_coord

        pursuer_coord = self.rng.choice(self.world.pursuer_start_coords)
        pursuer_state = np.zeros(PMBodyState.num_features(), dtype=np.float32)
        pursuer_state[:2] = pursuer_coord

        goal_coord = self.rng.choice(self.world.get_goal_coords(evader_coord))
        return PEState(
            evader_state,
            pursuer_state,
            evader_state[:2],
            pursuer_state[:2],
            np.array(goal_coord, dtype=np.float32),
            self.world.get_shortest_path_distance(evader_coord, goal_coord),
        )

    def sample_initial_obs(self, state: PEState) -> Dict[str, PEObs]:
        return self._get_obs(state)[0]

    def step(
        self, state: PEState, actions: Dict[str, PEAction]
    ) -> M.JointTimestep[PEState, PEObs]:
        clipped_actions = clip_actions(actions, self.action_spaces)
        next_state = self._get_next_state(state, clipped_actions)
        obs, evader_seen = self._get_obs(next_state)

        rewards = self._get_reward(state, next_state, evader_seen)
        all_done = self._is_done(next_state, evader_seen)
        terminated = {i: all_done for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}
        info: Dict[str, Dict] = {i: {} for i in self.possible_agents}
        if all_done:
            for i, outcome in self._get_outcome(next_state, evader_seen).items():
                info[i]["outcome"] = outcome

        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _get_next_state(self, state: PEState, actions: Dict[str, PEAction]) -> PEState:
        evader_a = actions[str(self.EVADER_IDX)]
        pursuer_a = actions[str(self.PURSUER_IDX)]

        self.world.set_entity_state("pursuer", state.pursuer_state)
        self.world.set_entity_state("evader", state.evader_state)

        pursuer_angle = state.pursuer_state[2] + pursuer_a[0]
        pursuer_vel = self.world.linear_to_xy_velocity(pursuer_a[1], pursuer_angle)
        self.world.update_entity_state("pursuer", angle=pursuer_angle, vel=pursuer_vel)

        evader_angle = state.evader_state[2] + evader_a[0]
        evader_vel = self.world.linear_to_xy_velocity(evader_a[1], evader_angle)
        self.world.update_entity_state("evader", angle=evader_angle, vel=evader_vel)

        # simulate
        self.world.simulate(1.0 / 10, 10)

        pursuer_next_state = np.array(
            self.world.get_entity_state("pursuer"),
            dtype=np.float32,
        )
        evader_next_state = np.array(
            self.world.get_entity_state("evader"),
            dtype=np.float32,
        )
        evader_coord = (evader_next_state[0], evader_next_state[1])
        goal_coord = (state.evader_goal_coord[0], state.evader_goal_coord[1])
        min_sp_distance = min(
            state.min_goal_dist,
            self.world.get_shortest_path_distance(evader_coord, goal_coord),
        )

        return PEState(
            evader_next_state,
            pursuer_next_state,
            state.evader_start_coord,
            state.pursuer_start_coord,
            state.evader_goal_coord,
            min_sp_distance,
        )

    def _get_obs(self, state: PEState) -> Tuple[Dict[str, PEObs], bool]:
        evader_obs, _ = self._get_agent_obs(state, evader=True)
        pursuer_obs, evader_seen = self._get_agent_obs(state, evader=False)

        return {
            str(self.EVADER_IDX): evader_obs,
            str(self.PURSUER_IDX): pursuer_obs,
        }, evader_seen

    def _get_agent_obs(
        self,
        state: PEState,
        evader: bool,
    ) -> Tuple[np.ndarray, bool]:
        if evader:
            agent_pos = (
                state.evader_state[0],
                state.evader_state[1],
                state.evader_state[2],
            )
            opp_coord = (state.pursuer_state[0], state.pursuer_state[1])
        else:
            agent_pos = (
                state.pursuer_state[0],
                state.pursuer_state[1],
                state.pursuer_state[2],
            )
            opp_coord = (state.evader_state[0], state.evader_state[1])

        ray_dists, ray_col_type = self.world.check_collision_circular_rays(
            agent_pos,
            self.max_obs_distance,
            self.n_sensors,
            np.array([opp_coord]),
            include_blocks=True,
            check_walls=True,
            use_relative_angle=True,
            angle_bounds=(-self.fov / 2, self.fov / 2),
        )

        obs = np.full((self.obs_dim,), self.max_obs_distance, dtype=np.float32)
        # Can bucket NONE type colisions with block/border/wall collisions, since these
        # will have dist of obs_dist and hence not affect final obs
        obs_type_idx = np.where(ray_col_type == CollisionType.AGENT.value, 1, 0)
        flat_obs_idx = np.ravel_multi_index(
            (obs_type_idx, np.arange(self.n_sensors)), dims=(2, self.n_sensors)
        )
        obs[flat_obs_idx] = ray_dists

        other_agent_seen = any(c == CollisionType.AGENT.value for c in ray_col_type)
        dist = self.world.euclidean_dist(agent_pos, opp_coord)

        aux_obs_idx = self.sensor_obs_dim
        obs[aux_obs_idx] = dist <= self.HEARING_DIST
        obs[aux_obs_idx + 1 : aux_obs_idx + 3] = state.evader_start_coord
        obs[aux_obs_idx + 3 : aux_obs_idx + 5] = state.pursuer_start_coord

        if evader:
            obs[aux_obs_idx + 5 : aux_obs_idx + 7] = state.evader_goal_coord
        else:
            obs[aux_obs_idx + 5 : aux_obs_idx + 7] = [0, 0]

        return obs, other_agent_seen

    def _get_reward(
        self, state: PEState, next_state: PEState, evader_seen: bool
    ) -> Dict[str, float]:
        evader_reward = 0.0
        if self._use_progress_reward and next_state.min_goal_dist < state.min_goal_dist:
            evader_reward += self.R_PROGRESS

        if evader_seen or self.world.agents_collide(
            next_state.evader_state, next_state.pursuer_state
        ):
            evader_reward -= self.R_EVASION
        elif self.world.agents_collide(
            next_state.evader_state, next_state.evader_goal_coord
        ):
            evader_reward += self.R_EVASION

        if self._normalize_reward:
            evader_reward = self._get_normalized_reward(evader_reward)

        return {
            str(self.EVADER_IDX): evader_reward,
            str(self.PURSUER_IDX): -evader_reward,
        }

    def _is_done(self, state: PEState, evader_seen: bool) -> bool:
        # need to do explicit bool conversion as conditions can return np.bool
        return bool(
            evader_seen
            or self.world.agents_collide(state.evader_state, state.pursuer_state)
            or self.world.agents_collide(state.evader_state, state.evader_goal_coord)
        )

    def _get_outcome(self, state: PEState, evader_seen: bool) -> Dict[str, M.Outcome]:
        evader_id, pursuer_id = str(self.EVADER_IDX), str(self.PURSUER_IDX)
        if evader_seen or self.world.agents_collide(
            state.evader_state, state.pursuer_state
        ):
            return {evader_id: M.Outcome.LOSS, pursuer_id: M.Outcome.WIN}
        if self.world.agents_collide(state.evader_state, state.evader_goal_coord):
            return {evader_id: M.Outcome.WIN, pursuer_id: M.Outcome.LOSS}
        return {evader_id: M.Outcome.DRAW, pursuer_id: M.Outcome.DRAW}

    def _get_normalized_reward(self, reward: float) -> float:
        """Normalize reward in [-1, 1] interval."""
        diff = self._max_raw_return - self._min_raw_return
        return 2 * (reward - self._min_raw_return) / diff - 1


class PEWorld(SquareContinuousWorld):
    """A world for the Pursuit Evasion Problem.

    Arguments
    ---------
    size : int
        height and width of the world.
    blocked_coords : Set[Coord]
        coordinates of blocked cell, i.e. 1x1 areas that contain walls.
    goal_coords_map : Dict[FloatCoord, List[FloatCoord]]
        map from evader start coordinates to a list of possible evader goal coordinates.
    evader_start_coords : List[FloatCoord]
        list of possible evader start coordinates.
    pursuer_start_coords : List[FloatCoord]
        list of possible pursuer start coordinates.

    """

    def __init__(
        self,
        size: int,
        blocked_coords: Set[Coord],
        goal_coords_map: Dict[FloatCoord, List[FloatCoord]],
        evader_start_coords: List[FloatCoord],
        pursuer_start_coords: List[FloatCoord],
    ):
        interior_walls = generate_interior_walls(size, size, blocked_coords)
        super().__init__(
            size=size,
            blocks=None,
            interior_walls=interior_walls,
            agent_radius=0.4,
            border_thickness=0.01,
            enable_agent_collisions=True,
        )
        self._blocked_coords = blocked_coords
        self._goal_coords_map = goal_coords_map
        self.evader_start_coords = evader_start_coords
        self.pursuer_start_coords = pursuer_start_coords
        self.shortest_paths = self.get_all_shortest_paths(self.all_goal_coords)

    def copy(self) -> "PEWorld":
        world = PEWorld(
            size=int(self.size),
            blocked_coords=self.blocked_coords,
            goal_coords_map=self._goal_coords_map,
            evader_start_coords=self.evader_start_coords,
            pursuer_start_coords=self.pursuer_start_coords,
        )
        for id, (body, shape) in self.entities.items():
            # make copies of each entity, and ensure the copies are linked correctly
            # and added to the new world and world space
            body = body.copy()
            shape = shape.copy()
            shape.body = body
            world.space.add(body, shape)
            world.entities[id] = (body, shape)
        return world

    @property
    def all_goal_coords(self) -> List[FloatCoord]:
        """The list of all evader goal locations."""
        all_locs = set()
        for v in self._goal_coords_map.values():
            all_locs.update(v)
        return list(all_locs)

    def get_goal_coords(self, evader_start_coord: FloatCoord) -> List[FloatCoord]:
        """Get list of possible evader goal coords for given start coords."""
        return self._goal_coords_map[evader_start_coord]

    def get_shortest_path_distance(self, coord: FloatCoord, dest: FloatCoord) -> int:
        """Get the shortest path distance from coord to destination."""
        coord_c = self.convert_to_coord(coord)
        dest_c = self.convert_to_coord(dest)
        return int(self.shortest_paths[dest_c][coord_c])

    def get_max_shortest_path_distance(self) -> int:
        """Get max shortest path distance between any start and goal coords."""
        max_dist = 0
        for start_coord, dest_coords in self._goal_coords_map.items():
            sp_dist = max(
                [
                    self.get_shortest_path_distance(start_coord, dest)
                    for dest in dest_coords
                ]
            )
            max_dist = max(max_dist, int(sp_dist))
        return max_dist


def get_8x8_world() -> PEWorld:
    """Generate the 8-by-8 PE world layout."""
    ascii_map = (
        "# 9 # 8#\n"
        "  #    #\n"
        " ### # 7\n"
        "  6   # \n"
        " ##5# # \n"
        "   #    \n"
        "#    #2 \n"
        " 0#1### \n"
    )
    return convert_map_to_world(ascii_map, 8, 8)


def get_16x16_world() -> PEWorld:
    """Generate the 16-by-16 PE world layout."""
    ascii_map = (
        "  ## ####### ###\n"
        "    9##    8 ###\n"
        "##       ### ###\n"
        "## ## ##      ##\n"
        "## ## ##   ## ##\n"
        "#  #  ##   #  7#\n"
        "# ## ###   # # #\n"
        "  #  6       # #\n"
        "#   ## ##  ##  #\n"
        "  #  #5##  #   #\n"
        " ###   #     # #\n"
        "     #   ###   #\n"
        " # # ### ##   ##\n"
        "0# #     #     #\n"
        "   ## ##   ##2  \n"
        "##     1 #####  \n"
    )
    return convert_map_to_world(ascii_map, 16, 16)


def get_32x32_world() -> PEWorld:
    """Generate the 32-by-32 PE world layout."""
    ascii_map = (
        "          #############  #######\n"
        "   #####           ####  #######\n"
        "      ##  #######     8  #######\n"
        " ###     9#               ######\n"
        " ###             #### ##  ######\n"
        " ##   #####  ##  #### ##   #####\n"
        " ##   #####  ##  ##        #####\n"
        "      ##### ####      ###  #####\n"
        " ### ###### ####      ###   ####\n"
        " ### #####  ####      ####  ####\n"
        " ##  ##### #####      ##### 7###\n"
        " ##  ####  #####       ####  ###\n"
        " #  #####  #####       #     ###\n"
        " #  ##### ######       # ##  ###\n"
        " #       6             # ##  ###\n"
        "    #       #   ##       ###  ##\n"
        "     #   ####  ###     ####   ##\n"
        "#### #   ####  ###    ####    ##\n"
        "#### #   ### 5####   ####     ##\n"
        "####  #      ####     ##  #   ##\n"
        "##### #      ####        ##   ##\n"
        "#                        #  #  #\n"
        "        ####        ##      #  #\n"
        "  #####  ####     #### ## #    #\n"
        "#   ####  ###     ####    # 2#  \n"
        "### ####  ##       ####     ####\n"
        "###       #  ##   1##        ###\n"
        "      #      ####       #       \n"
        " 0    ###### ######   ####      \n"
        "  ### ###### #        #####     \n"
        "# ### ####      ############    \n"
        "#                               \n"
    )
    return convert_map_to_world(ascii_map, 32, 32)


def convert_map_to_world(
    ascii_map: str,
    height: int,
    width: int,
    block_symbol: str = "#",
    pursuer_start_symbols: Optional[Set[str]] = None,
    evader_start_symbols: Optional[Set[str]] = None,
    evader_goal_symbol_map: Optional[Dict] = None,
) -> PEWorld:
    """Generate PE world layout from ascii map.

    By default
    - 0, 1, 2, 7, 8, 9 are possible evader start and goal locations
    - 5, 6 are possible pursuer start locations
    - # are blocks

    The default maps are approximate discrete versions of the map layout from:
    Seaman et al 2018, 'Nested Reasoning About Autonomous Agents Using
    Probabilistic Programs'

    """
    row_strs = ascii_map.splitlines()
    assert len(row_strs) > 1
    assert all(len(row) == len(row_strs[0]) for row in row_strs)
    assert len(row_strs[0]) > 1

    if pursuer_start_symbols is None:
        pursuer_start_symbols = {"3", "4", "5", "6"}
    if evader_start_symbols is None:
        evader_start_symbols = {"0", "1", "2", "7", "8", "9"}
    if evader_goal_symbol_map is None:
        evader_goal_symbol_map = {
            "0": ["7", "8", "9"],
            "1": ["7", "8", "9"],
            "2": ["8", "9"],
            "7": ["0", "1"],
            "8": ["0", "1", "2"],
            "9": ["0", "1", "2"],
        }

    blocked_coords: Set[Coord] = set()
    evader_start_coords = []
    pursuer_start_coords = []
    evader_symbol_coord_map = {}
    for r, c in product(range(height), range(width)):
        coord = (c + 0.5, r + 0.5)
        symbol = row_strs[r][c]
        if symbol == block_symbol:
            blocked_coords.add((c, r))
        if symbol in pursuer_start_symbols:
            pursuer_start_coords.append(coord)
        elif symbol in evader_start_symbols:
            evader_start_coords.append(coord)
            evader_symbol_coord_map[symbol] = coord

    evader_goal_coords_map = {}
    for start_symbol, goal_symbols in evader_goal_symbol_map.items():
        start_coord = evader_symbol_coord_map[start_symbol]
        evader_goal_coords_map[start_coord] = [
            evader_symbol_coord_map[symbol] for symbol in goal_symbols
        ]

    return PEWorld(
        size=width,
        blocked_coords=blocked_coords,
        goal_coords_map=evader_goal_coords_map,
        evader_start_coords=evader_start_coords,
        pursuer_start_coords=pursuer_start_coords,
    )


# world_name: world_make_fn
SUPPORTED_WORLDS: Dict[str, Callable[[], PEWorld]] = {
    "8x8": get_8x8_world,
    "16x16": get_16x16_world,
    "32x32": get_32x32_world,
}
